(* outputs to C code *)

structure GenC :> GENERATOR =
struct

  structure Var = Variable
  open IL7

  structure StringSet = SplaySetFn(
    struct
      type ord_key = string
      val compare = String.compare
    end)

  val $ = Layout.str
  val % = Layout.mayAlign
  val & = Layout.seq
  val ? = Layout.align

  fun var v =
    let
      fun fixchar c =
        if Char.isAlphaNum c then c
        else #"_"
    in
      (implode (map fixchar (explode (Var.toString v))))
    end
  fun bracket l= &[$"[", l, $"]"]

  val hotcint_t = $"hotcint_t"
  val hotc_t = $"hotc_t"
  val hotc_tptr = $"hotc_t*"

  val max_str_size = 255

  val header = ?[$"#include <runtime.h>",
                  $"extern hotcint_t exception_tag_counter;", $""]


  fun value_mem_usage v = 0
  
  fun bnd_mem_usage b =
    case b of
      BndAlloc(_, i) => i + 1
    | BndLit (_, Literal.LitString s) => 2
    | BndUnOp(_, oper, v) =>
        let val vsusage = value_mem_usage v
        in  case oper of
              Literal.OpIntToStr => (max_str_size div 4) + 2 + vsusage
            | Literal.OpCharToStr => 2
            | _ => 0
        end
    | BndBinOp(_, v1, oper, v2) =>
      let val vsusage = (value_mem_usage v1) + (value_mem_usage v2)
      in  case oper of
            Literal.OpConcat => (max_str_size div 4) + 2 + vsusage
          | _ => vsusage
      end
    | BndProj(_, _, v) => (value_mem_usage v)
    | BndSet(_, _, v) => (value_mem_usage v)
    | _ => 0

  fun exp_mem_usage e =
    case e of
      ExSwitch(v, es) => foldl Int.max 1 (map exp_mem_usage es)
    | ExHalt(v) => value_mem_usage v
    | ExLet(b, e) => (bnd_mem_usage b) + (exp_mem_usage e)
    | ExApp(v1, v2) => (value_mem_usage v1) + (value_mem_usage v2)

  fun gen_function_header(f, x) =
    &[$"void* ", $("hotc_"^ f),
      (Layout.paren (%[ hotc_tptr, $x]))]

  fun gen_proj (field, v) = Layout.paren(&[Layout.paren v, $".", $field])

  fun gen_prototype (f, (x, e)) =
    &[gen_function_header(f, "arg"), $";"]

  fun gen_int i =
    if i < 0 then
      &[$"-", $(Int.toString(abs i))]
    else
      $(Int.toString i)

  fun gen_application (f, arg) = &[$"app", Layout.tuple[f, arg]]
  
  and gen_cast (t, v) = Layout.paren (%[Layout.paren t, v])

  and gen_literal l =
    case l of
      Literal.LitInt i =>
        ($"", &[$"box_int", Layout.paren (gen_int i)])
    | Literal.LitChar c =>
        ($"", &[$"box_char", Layout.paren (&[$"'", $(Char.toString c), $"'"])])
    | Literal.LitString s =>
      let
        val svar = Var.newNamed "strconst"
        val s = String.toString s
        val sconst = "\"" ^ s ^ "\""
        val attr = " "
        (* words needed for the string rounded up *)
        val strsize = Int.toString ((String.size s + 3) div 4)
      in
      (?[&[hotc_t, $" ", ($ o var) svar, $" = box_ptr(hotc_alloc(", $strsize,
                $", TRUE));"],
         &[$"strcpy", Layout.tuple2 (fn x => x, $)
                          (&[$"unbox_str(", ($ o var) svar, $")"],
                              sconst), $";" ]],
       ($ o var) svar)
      end

  and gen_value fvs v =
    case v of
      ValVar x =>
        if StringSet.member(fvs, var x)
        then
          ($"", &[$"box_fun", Layout.paren (&[$"hotc_", ($ o var) x])])
        else ($"", ($ o var) x)
    | ValLit l => gen_literal l

  and gen_exp fvs e =
    case e of
      ExHalt v =>
        let val (prev, vstr) = gen_value fvs v
        in ?[prev, &[$"hotc_halt", Layout.paren vstr, $";"]]
        end
    | ExApp(v1, v2) =>
        let val (prev1, vstr1) = gen_value fvs v1
            val (prev2, vstr2) = gen_value fvs v2
        in ?[prev1, prev2, &[gen_application(vstr1, vstr2), $";"]]
        end
    | ExSwitch(v, branches) =>
      let
        val numbr = length branches
        val (_, branches) =
          foldr (fn (e, (i,l)) => (i - 1, (i, e)::l)) (numbr, []) branches
        fun gen_case_branch (i, e) =
          ?[&[$"case ", $(Int.toString i), $":", $" {"],
            Layout.indent 2 (gen_exp fvs e),
            $"}"
           ]
        val (prev, vstr) = gen_value fvs v
      in
        ?[prev,
          &[$"switch", Layout.paren (&[$"unbox_int",
            Layout.paren (vstr)]), $" {"],
          Layout.indent 2 (?(map gen_case_branch branches)),
          $"}"
         ]
      end
    | ExLet(b, e) =>
      ?[gen_bind fvs b,
        gen_exp fvs e]

  and gen_bind fvs b =
    let
     fun gen_dec(x, field, rvalue) =
      ?[&[hotc_t, $" ", ($ o var) x, $";"],
        &[gen_proj(field, ($ o var) x), $" = ", rvalue, $";"]]
     fun gen_assn(x, rvalue) =
        &[hotc_t, $" ", ($ o var) x, $" = ", rvalue, $";"]
    in
      case b of
        BndLit(x, l) =>
          let
            val (prev, lstr) = gen_literal l
          in
            ?[prev, gen_assn(x, lstr)] 
          end
      | BndUnOp(x, oper, v) =>
        let val (prev, vstr) = gen_value fvs v
        in
        ?[prev,
          gen_assn(x, &[$(Literal.unop_to_string oper),
                        Layout.paren vstr])
         ]
        end
      | BndBinOp(x, v1, oper, v2) =>
        let val (prev1, vstr1) = gen_value fvs v1
            val (prev2, vstr2) = gen_value fvs v2
        in
          ?[prev1, prev2,
            gen_assn(x, &[$(Literal.binop_to_string oper),
                          Layout.paren
                          (&[ vstr1,
                              $", ",
                              vstr2])])]
       end
      | BndAlloc(x, i) =>
          gen_assn(x,
            &[$"box_ptr",
              Layout.paren (&[$"hotc_alloc",
                              Layout.tuple2($, $) (Int.toString i, "FALSE")])])
      | BndNewTag x => gen_assn(x, $"box_int(exception_tag_counter++)")
      | BndProj(x, i, v) =>
        let val (prev, vstr) = gen_value fvs v
        in
          ?[prev,
            &[hotc_t, $" ", ($ o var) x, $" = ", $"get",
              Layout.tuple2(fn x => x, $ o Int.toString) (vstr, i), $";"]
           ]
        end
      | BndSet(x, i, v) =>
        let val (prev, vstr) = gen_value fvs v
        in
          ?[prev,
            &[$"set",
              Layout.tuple3($ o var, $ o Int.toString, fn x => x) (x, i, vstr),
              $";"]]
        end
    end

  fun gen_function fvs (f, (x, e)) =
    let
      val usage = exp_mem_usage e
    in
    ?[$"",
      &[gen_function_header(f, "arg"), $" {"],
      Layout.indent 2
        (?[&[$"hotc_check",
              Layout.tuple2 ($, $) ("arg", Int.toString usage), $";"],
           &[hotc_t, $" ", $x, $" = *arg;"],
           gen_exp fvs e]),
      $"}"]
    end

  fun generate_program (fbnds, e) =
    let
      val fbnds = map (fn (f, (x, e)) => ( var f, (var x, e)))
                      fbnds
      val fvs = StringSet.addList(StringSet.empty, map #1 fbnds)
      val fbnds = ("mainfun", ("unused", e))::fbnds
      val protos = ?(map gen_prototype fbnds)
      val bodies = ?(map (gen_function fvs) fbnds)
    in
      (Layout.tostringex 77 (?[header, $"", protos, bodies])) ^ "\n"
    end
end
