(* transforms the internal language of the elaboration assignment
   into the full Harper-Stone IL
 *)
structure Injector :> INJECTOR =
struct
  type context = ILStatic.context

  exception Convert of string

  open IL1

  fun tyList c =
    let
      val a = Variable.newNamed "list"
    in
      ConProj(1, ConMu(a, kndExp(KndType, 1),
        ConProd[
          TySum([(LabSum "1", tyUnit),
                 (LabSum "2", tyProd[c, ConProj(1, ConVar a)])])]))
    end

  fun convert_label lab =
    case lab of
      IL.LabTycon s => LabTycon s
    | IL.LabTyvar s => LabTyvar s
    | IL.LabExp s => LabExp s
    | IL.LabModule s => LabModule s
    | IL.LabInternal s => LabInternal s
    | IL.LabStarred l => LabStarred (convert_label l)

  fun convert_kind G knd =
    case knd of
      IL.KndType => KndType
    | IL.KndPi(a, k1, k2) =>
        let val (a, k2) = IL.alpha_vary (a, k2) IL.replace_con_kind
            val G' = IL.DecConKind(a, k1)::G
        in KndPi(a, convert_kind G k1, convert_kind G' k2)
        end
    | IL.KndProd ks => KndProd (map (convert_kind G) ks)
    | IL.KndSing c => KndSing(convert_con G c)

  and convert_con G con =
    case con of
      IL.ConVar a => ConVar a
    | IL.ConEvar (ref ebnd) => convert_ebind G ebnd
    | IL.ConAbs(a, k, c) =>
        let val (a, c) = IL.alpha_vary (a, c) IL.replace_con_con
            val G' = IL.DecConKind(a, k)::G
        in ConAbs(a, convert_kind G k, convert_con G' c)
        end
    | IL.ConApp(c1, c2) => ConApp(convert_con G c1, convert_con G c2)
    | IL.ConProd cs => ConProd(map (convert_con G) cs)
    | IL.ConProj(i, c) => ConProj(i, convert_con G c)
    | IL.ConModProj(m, l) =>
      let
        val IL.SigStruct decs = ILStatic.synth_mod_sig G m
        val labs = map (fn (x, d) => convert_label x) decs
      in
        ConModProj(convert_module G m, convert_label l, labs)
      end
    | IL.TyBool => tyBool
    | IL.TyUnit => tyUnit
    | IL.TyLit l => TyLit l
    | IL.TyProd(c1, c2) => tyProd [convert_con G c1, convert_con G c2]
    | IL.TyList c => tyList (convert_con G c)
    | IL.TyArrow (c1, c2) => TyArrow(convert_con G c1, convert_con G c2)
    | IL.TyForall(a, k, c) =>
        let val (a, c) = IL.alpha_vary (a, c) IL.replace_con_con
            val G' = IL.DecConKind(a, k)::G
        in TyForall(a, convert_kind G k, convert_con G' c)
        end

  and convert_ebind G ebnd =
    case ebnd of
      IL.EvFree _ => raise Convert
                            "Can't convert programs with free type variables."
    | IL.EvBound c => convert_con G c

  and convert_exp G exp =
    case exp of
      IL.ExVar x => ExVar x
    | IL.ExUnit => exUnit
    | IL.ExBool b => exBool b
    | IL.ExIf(e1, e2, e3) =>
        let val cres = ILStatic.synth_exp_con G e2
            val e1' = convert_exp G e1
            val e2' = convert_exp G e2
            val e3' = convert_exp G e3
        in
          exIf (convert_con G cres) (e1', e2', e3')
        end
    | IL.ExLit l => ExLit l
    | IL.ExUnOp(oper, e) => ExUnOp(oper, convert_exp G e)
    | IL.ExBinOp(e1, oper, e2) =>
        ExBinOp(convert_exp G e1, oper, convert_exp G e2)
    | IL.ExPair(e1, e2) =>
        ExRecord([(LabRecord "1", convert_exp G e1),
                  (LabRecord "2", convert_exp G e2)])
    | IL.ExProj1 e =>
        ExRecordProj(LabRecord "1",
                     convert_exp G e,
                     [LabRecord "1", LabRecord "2"])
    | IL.ExProj2 e =>
        ExRecordProj(LabRecord "2",
                     convert_exp G e,
                     [LabRecord "1", LabRecord "2"])
    | IL.ExNil c =>
      let val c' = convert_con G c
          val clist = tyList c'
      in
        ExRoll(clist,
               ExSumInj(TySum[(LabSum "1", tyUnit),
                              (LabSum "2", tyProd [c', clist])],
                        LabSum "1",
                        exUnit,
                        [LabSum "1", LabSum "2"]))
      end
    | IL.ExCons (e1, e2) =>
        let
          val c = convert_con G (ILStatic.synth_exp_con G e1)
          val clist = tyList c
        in
          ExRoll(clist,
            ExSumInj(TySum[(LabSum "1", tyUnit),
                           (LabSum "2", tyProd [c, clist])],
                     LabSum "2",
                     exProd [convert_exp G e1, convert_exp G e2],
                     [LabSum "1", LabSum "2"]))
        end
    | IL.ExCase (e, e1, (h, t, e2)) =>
      let
        val c =
          case ILStatic.synth_exp_con G e of
            IL.TyList c => c
          | _ => raise Convert "Case at non-list type"
        val c' = convert_con G c
        val cres = ILStatic.synth_exp_con G e1
        val (h, e2) = IL.alpha_vary (h, e2) IL.replace_exp_exp
        val (t, e2) = IL.alpha_vary (t, e2) IL.replace_exp_exp
        val cres' = convert_con G cres
        val e' = convert_exp G e
        val e1' = convert_exp G e1
        val G' = (IL.DecExpCon(h, c))::(IL.DecExpCon(t, IL.TyList c))::G
        val e2' = convert_exp G' e2
        val x = Variable.new()
        val cases =
          [(LabSum "1", exLam(x, tyUnit, cres', e1')),
           (LabSum "2",
              exLam(x, tyProd [c', tyList c'],
                    cres',
                    ExLet(h, ExRecordProj(LabRecord "1", ExVar x,
                            [LabRecord "1", LabRecord "2"]),
                    ExLet(t, ExRecordProj(LabRecord "2", ExVar x,
                            [LabRecord "1", LabRecord "2"]),
                      e2'))))
          ]
      in
        ExCase(cres', ExUnroll e', cases)
      end
    | IL.ExFix(f, x, c1, c2, e) =>
        let val (f, e) = IL.alpha_vary (f, e) IL.replace_exp_exp
            val (x, e) = IL.alpha_vary (x, e) IL.replace_exp_exp
            val c1' = convert_con G c1
            val c2' = convert_con G c2
            val G' =
              (IL.DecExpCon(f, IL.TyArrow(c1, c2)))::(IL.DecExpCon(x, c1))::G
            val e' = convert_exp G' e
        in ExRecordProj(
            LabRecord "1", ExFix[(f, x, c1', c2', e')], [LabRecord "1"])
        end
    | IL.ExApp(e1, e2) => ExApp(convert_exp G e1, convert_exp G e2)
    | IL.ExConAbs(a, k, e) =>
        let val (a, e) = IL.alpha_vary (a, e) IL.replace_con_exp
        in ExConAbs(a, convert_kind G k, convert_exp G e)
        end
    | IL.ExConApp(e, c) => ExConApp(convert_exp G e, convert_con G c)
    | IL.ExModLet(s, m, e) =>
        let val (s, e) = IL.alpha_vary (s, e) IL.replace_mod_exp
            val sg = ILStatic.synth_mod_sig G m
            val G' = IL.DecModSig(s, sg)::G
            (* we assume that s \notin c since the baby elaborator should
              reject programs that do *)
            val cres = ILStatic.synth_exp_con G' e
        in ExModLet(s, convert_module G m,
                    convert_exp G' e, convert_con G' cres)
        end
    | IL.ExModProj(m, l) =>
      let
        val IL.SigStruct decs = ILStatic.synth_mod_sig G m
        val labs = map (fn (x, d) => convert_label x) decs
      in
        ExModProj(convert_module G m, convert_label l, labs)
      end

  and convert_module G module =
    case module of
      IL.ModVar s => ModVar s
    | IL.ModStruct sbnds =>
      let
      in
        ModStruct (convert_sbnds G sbnds)
      end
    | IL.ModModProj(m, l) =>
      let
        val IL.SigStruct decs = ILStatic.synth_mod_sig G m
        val labs = map (fn (x, d) => convert_label x) decs
      in
        ModModProj(convert_module G m, convert_label l, labs)
      end
    | IL.ModConstrain(m, sg) =>
          ModConstrain(convert_module G m, convert_signat G sg)

    and convert_sbnds G sbnds =
      let
        fun convert_sbnds' G [] x = x
          | convert_sbnds' G ((l, bnd)::sbnds) accum =
              (case bnd of
                IL.BndExp(x, e) =>
                  let
                    val (x, sbnds) =
                      IL.alpha_vary (x, sbnds) IL.replace_exp_sbnds
                    val cres = ILStatic.synth_exp_con G e
                    val G' = IL.DecExpCon(x, cres)::G
                    val accum' =
                      (convert_label l, BndExp(x, convert_exp G e))::accum
                  in
                    convert_sbnds' G' sbnds accum'
                  end
               | IL.BndCon(a, c) =>
                  let
                    val (a, sbnds) =
                      IL.alpha_vary (a, sbnds) IL.replace_con_sbnds
                    val kres = ILStatic.synth_con_kind G c
                    val G' = IL.DecConKind(a, kres)::G
                    val accum' =
                      (convert_label l, BndCon(a, convert_con G c))::accum
                  in
                    convert_sbnds' G' sbnds accum'
                  end
               | IL.BndMod(s, m) =>
                  let
                    val (s, sbnds) =
                      IL.alpha_vary (s, sbnds) IL.replace_mod_sbnds
                    val sgres = ILStatic.synth_mod_sig G m
                    val G' = IL.DecModSig(s, sgres)::G
                    val accum' =
                      (convert_label l, BndMod(s, convert_module G m))::accum
                  in
                    convert_sbnds' G' sbnds accum'
                  end)
      in
         rev (convert_sbnds' G sbnds [])
      end

  and convert_signat G (IL.SigStruct sdecs) =
      let
        fun convert_sdecs' G [] x = x
          | convert_sdecs' G ((l, dec)::sdecs) accum =
              (case dec of
                IL.DecExpCon(x, c) =>
                  let
                    val (x, sdecs) =
                      IL.alpha_vary (x, sdecs) IL.replace_exp_sdecs
                    val G' = IL.DecExpCon(x, c)::G
                    val newdec =
                      (convert_label l, DecExpCon(x, convert_con G c))
                  in
                    convert_sdecs' G' sdecs (newdec::accum)
                  end
               | IL.DecConKind(a, k) =>
                  let
                    val (a, sdecs) =
                      IL.alpha_vary (a, sdecs) IL.replace_con_sdecs
                    val G' = IL.DecConKind(a, k)::G
                    val newdec =
                      (convert_label l, DecConKind(a, convert_kind G k))
                  in
                    convert_sdecs' G' sdecs (newdec::accum)
                  end
               | IL.DecModSig(s, sg) =>
                  let
                    val (s, sdecs) =
                      IL.alpha_vary (s, sdecs) IL.replace_mod_sdecs
                    val G' = IL.DecModSig(s, sg)::G
                    val newdec =
                        (convert_label l, DecModSig(s, convert_signat G sg))
                  in
                    convert_sdecs' G' sdecs (newdec::accum)
                  end)
      in
        SigStruct (rev (convert_sdecs' G sdecs []))
      end
end
