structure ELPrint =
struct
  open EL
  
  structure Lit = Literal
  
  val $ = Layout.str
  val % = Layout.mayAlign
  val & = Layout.seq

  fun longid_to_layout (ids, id) =
      &(Layout.separate (map $ (ids @ [id]), "."))

  fun atomic_ty_to_layout ty =
      case ty of
          TyVar id => $id
        | _ => Layout.paren (ty_to_layout ty)

  and app_ty_to_layout ty =
      case ty of
          TyBool => $"bool"
        | TyLit (Literal.LitTyInt) => $"int"
        | TyLit (Literal.LitTyChar) => $"char"
        | TyLit (Literal.LitTyString) => $"string"
        | TyUnit => $"unit"
        | TyList tyarg => &[app_ty_to_layout tyarg, $" list"]
        | TyCon ([], longid) => longid_to_layout longid
        | TyCon ([tyarg], longid) =>
              &[app_ty_to_layout tyarg, $" ", longid_to_layout longid]
        | TyCon (tys, longid) =>
              let val tyls = map ty_to_layout tys
                  val idl = longid_to_layout longid
              in
                  &[Layout.tuple tyls, $" ", idl]
              end
        | _ => atomic_ty_to_layout ty

  and prod_ty_to_layout ty =
      case ty of
          TyProd (ty1, ty2) =>
              %[app_ty_to_layout ty1,
                &[$"* ", app_ty_to_layout ty2]]
        | _ => app_ty_to_layout ty

  and arrow_ty_to_layout ty =
      case ty of
          TyArrow (ty1, ty2) =>
              %[prod_ty_to_layout ty1,
                &[$"-> ", arrow_ty_to_layout ty2]]
        | _ => prod_ty_to_layout ty

  and ty_to_layout ty = arrow_ty_to_layout ty

  fun atomic_exp_to_layout exp =
      case exp of
          ExVar longid => longid_to_layout longid
        | ExUnit => $"()"
        | ExBool b => $(if b then "true" else "false")
        | ExLit lit =>
                (case lit of
                    Literal.LitInt i => $(Int.toString i)
                  | Literal.LitChar c => $("'" ^ Char.toString c ^ "'")
                  | Literal.LitString s => $("\"" ^ s ^ "\""))
        | ExPair (e1, e2) => Layout.tuple [exp_to_layout e1,
                                           exp_to_layout e2]
        | ExNil => $"nil"
        | ExLet (bindings, e) =>
              Layout.align
              [ &[$"let ", bindings_to_layout bindings],
                $"in",
                Layout.indent 4 (exp_to_layout e),
                $"end" ]
        | _ => Layout.paren (exp_to_layout exp)

  and app_exp_to_layout exp =
      case exp of
          ExApp (e1, e2) => %[app_exp_to_layout e1,
                              atomic_exp_to_layout e2]
        | ExProj1 e => %[$"#1", atomic_exp_to_layout e]
        | ExProj2 e => %[$"#2", atomic_exp_to_layout e]
        | ExUnOp(oper, e) => %[$(Lit.unop_to_string oper),
                              atomic_exp_to_layout e]
        | _ => atomic_exp_to_layout exp

  and times_exp_to_layout exp =
      case exp of
          ExBinOp (e1, Lit.OpTimes, e2) =>
              %[times_exp_to_layout e1,
                &[$"* ", app_exp_to_layout e2]]
        | ExBinOp (e1, Lit.OpConcat, e2) =>
              %[times_exp_to_layout e1,
                &[$"^ ", app_exp_to_layout e2]]
        | _ => app_exp_to_layout exp
  
  and plus_minus_exp_to_layout exp =
      case exp of
          ExBinOp (e1, Lit.OpPlus, e2) =>
              %[plus_minus_exp_to_layout e1,
                &[$"+ ", times_exp_to_layout e2]]
        | ExBinOp (e1, Lit.OpMinus, e2) =>
              %[plus_minus_exp_to_layout e1,
                &[$"- ", times_exp_to_layout e2]]
        | _ => times_exp_to_layout exp

  and cons_exp_to_layout exp =
      case exp of
          ExCons (e1, e2) =>
              %[plus_minus_exp_to_layout e1,
                &[$":: ", cons_exp_to_layout e2]]
        | _ => plus_minus_exp_to_layout exp

  and equals_exp_to_layout exp =
      case exp of
          ExBinOp (e1, Lit.OpEquals, e2) =>
              %[cons_exp_to_layout e1,
                &[$"= ", cons_exp_to_layout e2]]
        | _ => cons_exp_to_layout exp

  and constrain_exp_to_layout exp =
      case exp of
          ExConstrain (e, ty) =>
              %[equals_exp_to_layout e,
                &[$": ", ty_to_layout ty]]
        | _ => equals_exp_to_layout exp

  and exp_to_layout exp =
      case exp of
          ExIf (e1, e2, e3) =>
              Layout.align
              [ &[$"if ", exp_to_layout e1, $" then"],
                Layout.indent 4 (exp_to_layout e2),
                $"else",
                Layout.indent 4 (exp_to_layout e3) ]
        | ExCase (e, e_nil, (x, xs, e_cons)) =>
              Layout.align
              [ &[$"case ", exp_to_layout e, $" of"],
                Layout.indent 4 (&[$"nil => ", exp_to_layout e_nil]),
                Layout.indent 2 (&[$"| ", $x, $" :: ", $xs, $" => ",
                                   exp_to_layout e_cons]) ]
        | ExFn (x, e) =>
              %[ &[$"fn ", $x, $" =>"],
                 Layout.indent 4 (exp_to_layout e) ]
        | _ => constrain_exp_to_layout exp

  and bindings_to_layout bindings =
      Layout.align (map binding_to_layout bindings)

  and app_ty_to_layout (tyvars, id) =
      case tyvars of
          [] => $id
        | [tyvar] => &[$tyvar, $" ", $id]
        | _ => &[Layout.tuple (map $ tyvars), $" ", $id]

  and binding_to_layout binding =
      case binding of
          BndVal (id, e) =>
              %[ &[$"val ", $id, $" ="],
                 Layout.indent 4 (exp_to_layout e) ]
        | BndFun (f, x, e) =>
              %[ &[$"fun ", $f, $" ", $x, $" ="],
                 Layout.indent 4 (exp_to_layout e) ]
        | BndType (tyvars, id, ty) =>
              %[ &[$"type ", app_ty_to_layout (tyvars, id), $" ="],
                 Layout.indent 4 (ty_to_layout ty) ]
        | BndOpen longid =>
              &[$"open ", longid_to_layout longid]
        | BndLocal (bnds1, bnds2) =>
              Layout.align
              [$"local",
               Layout.indent 4 (bindings_to_layout bnds1),
               $"in",
               Layout.indent 4 (bindings_to_layout bnds2),
               $"end"]
        | BndStructure (id, strexp) =>
              %[ &[$"structure ", $id, $" ="],
                 Layout.indent 4 (strexp_to_layout strexp)]

  and strexp_to_layout strexp =
      case strexp of
          StrStruct bindings =>
              %[$"struct",
                Layout.indent 4 (bindings_to_layout bindings),
                $"end"]
        | StrVar longid => longid_to_layout longid
        | StrSealOpaque (strexp, sigexp) =>
              %[ strexp_to_layout strexp,
                 Layout.indent 2 (&[$":> ", sigexp_to_layout sigexp]) ]
        | StrSealTransparent (strexp, sigexp) =>
              %[ strexp_to_layout strexp,
                 Layout.indent 2 (&[$": ", sigexp_to_layout sigexp]) ]
              (*
              &[strexp_to_layout strexp,
                $" : ", sigexp_to_layout sigexp]
              *)
        | StrLet (bindings, strexp) =>
              Layout.align
              [ &[$"let ", bindings_to_layout bindings],
                $"in",
                Layout.indent 4 (strexp_to_layout strexp),
                $"end" ]

  and sigexp_to_layout sigexp =
      case sigexp of
          SigSig specs =>
              %[$"sig",
                Layout.indent 4 (specs_to_layout specs),
                $"end"]

  and specs_to_layout specs =
      Layout.align (map spec_to_layout specs)

  and spec_to_layout spec =
      case spec of
          SpecVal (id, ty) =>
              %[ &[$"val ", $id, $" :"],
                 Layout.indent 4 (ty_to_layout ty) ]
        | SpecType (tyvars, id, NONE) =>
              &[$"type ", app_ty_to_layout (tyvars, id)]
        | SpecType (tyvars, id, SOME ty) =>
              %[ &[$"type ", app_ty_to_layout (tyvars, id), $" ="],
                 Layout.indent 4 (ty_to_layout ty) ]
        | SpecStructure (id, sigexp) =>
              %[ &[$"structure ", $id, $" :"],
                 Layout.indent 4 (sigexp_to_layout sigexp)]

  val program_to_layout = bindings_to_layout
end
