structure ILPrint =
struct
  open IL

  structure Lit = Literal

  val $ = Layout.str
  val % = Layout.mayAlign
  val & = Layout.seq

  fun label lab = $(string_of_label lab)
  fun var v = $(Variable.toString v)

  fun evar_normalize_con con =
      map_con (fn cx => ConVar cx)
              (fn ebr => case !ebr of
                              EvFree _ => ConEvar ebr
                            | EvBound c => evar_normalize_con c)
              (fn (cx, c) => (cx, evar_normalize_con c))
              evar_normalize_kind
              evar_normalize_module
              con

  and evar_normalize_kind knd =
      map_kind (fn (cx, k) => (cx, evar_normalize_kind k))
               evar_normalize_con
               knd

  and evar_normalize_module module =
      map_module (fn mx => ModVar mx)
                 evar_normalize_sbnds
                 evar_normalize_signat
                 module

  and evar_normalize_sbnds sbnds =
      map_sbnds (fn (ex, sbs) => (ex, evar_normalize_sbnds sbs))
                (fn (cx, sbs) => (cx, evar_normalize_sbnds sbs))
                (fn (mx, sbs) => (mx, evar_normalize_sbnds sbs))
                evar_normalize_exp
                evar_normalize_con
                evar_normalize_module
                sbnds

  and evar_normalize_signat signat =
      map_signat evar_normalize_sdecs
                 signat

  and evar_normalize_sdecs sdecs =
      map_sdecs (fn (ex, sds) => (ex, evar_normalize_sdecs sds))
                (fn (cx, sds) => (cx, evar_normalize_sdecs sds))
                (fn (mx, sds) => (mx, evar_normalize_sdecs sds))
                evar_normalize_con
                evar_normalize_kind
                evar_normalize_signat
                sdecs

  and evar_normalize_exp exp =
      map_exp (fn ex => ExVar ex)
              (fn (cx, e) => (cx, evar_normalize_exp e))
              (fn (mx, e) => (mx, evar_normalize_exp e))
              evar_normalize_con
              evar_normalize_kind
              evar_normalize_module
              exp

  local
      val ebrs : (ebind ref * int) list ref = ref []
      val r = ref 0
  in
      fun ebr_to_string ebr =
          (* simple memoization routine to assign ints to ebind refs *)
          let val n =
              case ListUtil.Alist.find op= (!ebrs) ebr of
                  NONE =>
                      let val n = !r
                      in
                          r := n + 1;
                          ebrs := (ebr, n) :: !ebrs;
                          n
                      end
                | SOME n => n
          in
              case !ebr of
                  EvFree (condeps, moddeps) =>
                      let fun concatWith s l =
                              case l of
                                  [] => ""
                                | [x] => x
                                | x::xs => x ^ s ^ concatWith s xs
                          val cdl =
                          ConVarSet.foldl
                              (fn (cx, l) => Variable.toString cx :: l)
                              []
                              condeps
                          val mdl =
                          ModVarSet.foldl
                              (fn (mx, l) => Variable.toString mx :: l)
                              []
                              moddeps
                          val cds = concatWith "," cdl
                          val mds = concatWith "," mdl
                      in
                          "E" ^ Int.toString n ^ "["
                          ^ "{" ^ cds ^ "};"
                          ^ "{" ^ mds ^ "}"
                          ^ "]"
                      end
                | _ => raise Fail "impossible"
          end
  end


  (*** kinds ***)

  fun kind_to_layout knd =
      case knd of
          KndType => $"Type"
        | KndPi (cx, k, k') =>
              %[ &[$"Pi ", var cx, $":", kind_to_layout k, $"."],
                 kind_to_layout k']
        | KndProd ks =>
              Layout.sequence ("<", ">", " *")
                  (map kind_to_layout ks)
        | KndSing c => &[$"S(", con_to_layout c, $")"]


  (*** constructors ***)

  and atomic_con_to_layout con =
      case con of
          ConVar cx => var cx
        | TyBool => $"bool"
        | TyLit (Literal.LitTyInt) => $"int"
        | TyLit (Literal.LitTyChar) => $"char"
        | TyLit (Literal.LitTyString) => $"string"
        | TyUnit => $"unit"
        | ConEvar ebr =>
              (case !ebr of
                  EvFree _ => $(ebr_to_string ebr)
                | EvBound _ => raise Fail "unnormalized evar!")
        | ConProd cs =>
              Layout.sequence ("<", ">", ",")
                  (map con_to_layout cs)
        | ConModProj (m, lab) =>
              &[module_to_layout m, $".", label lab]
        | _ => Layout.paren (con_to_layout con)

  and app_con_to_layout con =
      case con of
          TyList conarg =>
              &[$"list ", atomic_con_to_layout conarg]
        | ConApp (c1, c2) =>
              %[app_con_to_layout c1,
                atomic_con_to_layout c2]
        | ConProj (i, c) =>
              %[ &[$"#", $(Int.toString i)],
                 atomic_con_to_layout c ]
        | _ => atomic_con_to_layout con

  and prod_con_to_layout con =
      case con of
          TyProd (con1, con2) =>
              %[app_con_to_layout con1,
                &[$"* ", app_con_to_layout con2]]
        | _ => app_con_to_layout con

  and arrow_con_to_layout con =
      case con of
          TyArrow (con1, con2) =>
              %[prod_con_to_layout con1,
                &[$"-> ", arrow_con_to_layout con2]]
        | _ => prod_con_to_layout con

  and con_to_layout con =
      (* first, eliminate bound evars *)
      let val con = evar_normalize_con con
      in
          case con of
              TyForall (cx, k, c) =>
                  %[ &[$"All ", var cx, $":", kind_to_layout k, $"."],
                     con_to_layout c ]
            | ConAbs (cx, k, c) =>
                  %[ &[$"\\ ", var cx, $":", kind_to_layout k, $"."],
                     con_to_layout c ]
            | _ => arrow_con_to_layout con
      end


  (*** expressions ***)

  and atomic_exp_to_layout exp =
      case exp of
          ExVar x => var x
        | ExUnit => $"()"
        | ExBool b => $(if b then "true" else "false")
        | ExLit lit =>
                (case lit of
                    Literal.LitInt i => $(Int.toString i)
                  | Literal.LitChar c => $("'" ^ Char.toString c ^ "'")
                  | Literal.LitString s => $("\"" ^ s ^ "\""))
        | ExPair (e1, e2) => Layout.tuple [exp_to_layout e1,
                                           exp_to_layout e2]
        | ExNil c => &[$"nil [", con_to_layout c, $"]"]
        | ExModProj (m, lab) =>
              &[module_to_layout m, $".", label lab]
        | ExModLet (mx, m, e) =>
              Layout.align
              [ %[ &[$"let ", var mx, $" ="],
                   Layout.indent 4 (module_to_layout m) ],
                $"in",
                Layout.indent 4 (exp_to_layout e),
                $"end" ]
        | _ => Layout.paren (exp_to_layout exp)

  and app_exp_to_layout exp =
      case exp of
          ExApp (e1, e2) => %[app_exp_to_layout e1,
                              atomic_exp_to_layout e2]
        | ExConApp (e, c) => %[app_exp_to_layout e,
                              &[$"[", con_to_layout c, $"]"]]
        | ExProj1 e => %[$"#1", atomic_exp_to_layout e]
        | ExProj2 e => %[$"#2", atomic_exp_to_layout e]
        | ExUnOp(oper, e) => %[$(Lit.unop_to_string oper),
                              atomic_exp_to_layout e]
        | _ => atomic_exp_to_layout exp

  and times_exp_to_layout exp =
      case exp of
          ExBinOp (e1, Lit.OpTimes, e2) =>
              %[times_exp_to_layout e1,
                &[$"* ", app_exp_to_layout e2]]
        | ExBinOp (e1, Lit.OpConcat, e2) =>
              %[times_exp_to_layout e1,
                &[$"^ ", app_exp_to_layout e2]]
        | _ => app_exp_to_layout exp
  
  and plus_minus_exp_to_layout exp =
      case exp of
          ExBinOp (e1, Lit.OpPlus, e2) =>
              %[plus_minus_exp_to_layout e1,
                &[$"+ ", times_exp_to_layout e2]]
        | ExBinOp (e1, Lit.OpMinus, e2) =>
              %[plus_minus_exp_to_layout e1,
                &[$"- ", times_exp_to_layout e2]]
        | _ => times_exp_to_layout exp

  and cons_exp_to_layout exp =
      case exp of
          ExCons (e1, e2) =>
              %[plus_minus_exp_to_layout e1,
                &[$":: ", cons_exp_to_layout e2]]
        | _ => plus_minus_exp_to_layout exp

  and equals_exp_to_layout exp =
      case exp of
          ExBinOp (e1, Lit.OpEquals, e2) =>
              %[cons_exp_to_layout e1,
                &[$"= ", cons_exp_to_layout e2]]
        | _ => cons_exp_to_layout exp

  and exp_to_layout exp =
      case exp of
          ExIf (e1, e2, e3) =>
              Layout.align
              [ &[$"if ", exp_to_layout e1, $" then"],
                Layout.indent 4 (exp_to_layout e2),
                $"else",
                Layout.indent 4 (exp_to_layout e3) ]
        | ExCase (e, e_nil, (x, xs, e_cons)) =>
              Layout.align
              [ &[$"case ", exp_to_layout e, $" of"],
                Layout.indent 4 (&[$"nil => ", exp_to_layout e_nil]),
                Layout.indent 2 (&[$"| ", var x, $" :: ", var xs, $" => ",
                                   exp_to_layout e_cons]) ]
        | ExFix (f, x, c1, c2, e) =>
              (* XXX which looks better? *)
              (*
              %[ &[$"fix ",
                   %[ &[var f, $" (", var x, $" : ", con_to_layout c1, $")"],
                      &[$": ", con_to_layout c2, $" =>"] ] ],
                 Layout.indent 4 (exp_to_layout e) ]
              *)
              %[ &[$"fix ", var f, $" (", var x, $" : ", con_to_layout c1, $")"],
                 Layout.indent 2 (&[$": ", con_to_layout c2, $" =>"]),
                 Layout.indent 4 (exp_to_layout e) ]
        | ExConAbs (cx, k, e) =>
              %[ &[$"/\\ ", var cx, $":", kind_to_layout k, $"."],
                 exp_to_layout e]
        | _ => equals_exp_to_layout exp


  (*** modules ***)

  and module_to_layout module =
      case module of
          ModVar mx => var mx
        | ModStruct sbnds =>
              &[ $"[", sbnds_to_layout sbnds, $"]" ]
        | ModModProj (m, lab) =>
              &[module_to_layout m, $".", label lab]
        | ModConstrain (m, sg) =>
              %[module_to_layout m,
                &[$":> ", signat_to_layout sg]]
      
  and sbnds_to_layout sbnds =
      %(Layout.separateRight (map sbnd_to_layout sbnds, ","))

  and sbnd_to_layout sbnd =
      case sbnd of
          (lab, BndExp (ex, e)) =>
              (*
              &[label lab, $" ",
                %[ &[$"|> ", var ex],
                   Layout.indent 1 (&[$"= ", exp_to_layout e]) ] ]
              *)
              %[ &[label lab, $" |> ", var ex, $" ="],
                 Layout.indent 4 (exp_to_layout e) ]
        | (lab, BndCon (cx, c)) =>
              %[ &[label lab, $" |> ", var cx, $" ="],
                 Layout.indent 4 (con_to_layout c) ]
        | (lab, BndMod (mx, m)) =>
              %[ &[label lab, $" |> ", var mx, $" ="],
                 Layout.indent 4 (module_to_layout m) ]

  and signat_to_layout signat =
      case signat of
          SigStruct sdecs =>
              &[ $"[", sdecs_to_layout sdecs, $"]" ]

  and sdecs_to_layout sdecs =
      %(Layout.separateRight (map sdec_to_layout sdecs, ","))
  
  and sdec_to_layout sdec =
      case sdec of
          (lab, DecExpCon (ex, c)) =>
              %[ &[label lab, $" |> ", var ex, $" :"],
                 Layout.indent 4 (con_to_layout c) ]
        | (lab, DecConKind (cx, k)) =>
              %[ &[label lab, $" |> ", var cx, $" :"],
                 Layout.indent 4 (kind_to_layout k) ]
        | (lab, DecModSig (mx, sg)) =>
              %[ &[label lab, $" |> ", var mx, $" :"],
                 Layout.indent 4 (signat_to_layout sg) ]

  fun ctx_to_layout ctx =
      &[ $"{", decs_to_layout (List.rev ctx), $"}" ]

  and decs_to_layout sdecs =
      %(Layout.separateRight (map dec_to_layout sdecs, ","))

  and dec_to_layout dec =
      case dec of
          DecExpCon (ex, c) =>
              %[ &[var ex, $" :"], Layout.indent 4 (con_to_layout c) ]
        | DecConKind (cx, k) =>
              %[ &[var cx, $" :"], Layout.indent 4 (kind_to_layout k) ]
        | DecModSig (mx, sg) =>
              %[ &[var mx, $" :"], Layout.indent 4 (signat_to_layout sg) ]

  val program_to_layout = sbnds_to_layout

  fun con_to_string con = Layout.tostring (con_to_layout con)
end
