structure IL1Print =
  struct
    val $ = Layout.str
    val % = Layout.mayAlign
    val & = Layout.seq

    open IL1

    fun label lab = $(string_of_label lab)
    fun var v = $(Variable.toString v)

    fun evar_normalize_con con =
        map_con (fn cx => ConVar cx)
                (fn ebr => case !ebr of
                                EvFree _ => ConEvar ebr
                              | EvBound c => evar_normalize_con c)
                (fn (cx, c) => (cx, evar_normalize_con c))
                evar_normalize_kind
                evar_normalize_module
                con

    and evar_normalize_kind knd =
        map_kind (fn (cx, k) => (cx, evar_normalize_kind k))
                 evar_normalize_con
                 knd

    and evar_normalize_module module =
        map_module (fn mx => ModVar mx)
                   evar_normalize_sbnds
                   (fn (mx, m) => (mx, evar_normalize_module m))
                   evar_normalize_signat
                   module

    and evar_normalize_sbnds sbnds =
        map_sbnds (fn (ex, sbs) => (ex, evar_normalize_sbnds sbs))
                  (fn (cx, sbs) => (cx, evar_normalize_sbnds sbs))
                  (fn (mx, sbs) => (mx, evar_normalize_sbnds sbs))
                  evar_normalize_exp
                  evar_normalize_con
                  evar_normalize_module
                  sbnds

    and evar_normalize_signat signat =
        map_signat (fn (mx, sg) => (mx, evar_normalize_signat sg))
                   evar_normalize_sdecs
                   signat

    and evar_normalize_sdecs sdecs =
        map_sdecs (fn (ex, sds) => (ex, evar_normalize_sdecs sds))
                  (fn (cx, sds) => (cx, evar_normalize_sdecs sds))
                  (fn (mx, sds) => (mx, evar_normalize_sdecs sds))
                  evar_normalize_con
                  evar_normalize_kind
                  evar_normalize_signat
                  sdecs

    and evar_normalize_exp exp =
        map_exp (fn ex => ExVar ex)
                evar_normalize_fbnds
                (fn (ex, e) => (ex, evar_normalize_exp e))
                (fn (cx, e) => (cx, evar_normalize_exp e))
                (fn (mx, e) => (mx, evar_normalize_exp e))
                evar_normalize_con
                evar_normalize_kind
                evar_normalize_module
                exp

    and evar_normalize_fbnds fbnds =
        map (fn (f, x, con, con', exp) =>
                (f, x, evar_normalize_con con,
                       evar_normalize_con con',
                       evar_normalize_exp exp))
            fbnds

    local
        val ebrs : (ebind ref * int) list ref = ref []
        val r = ref 0
    in
        fun ebr_to_string ebr =
            (* simple memoization routine to assign ints to ebind refs *)
            let val n =
                case ListUtil.Alist.find op= (!ebrs) ebr of
                    NONE =>
                        let val n = !r
                        in
                            r := n + 1;
                            ebrs := (ebr, n) :: !ebrs;
                            n
                        end
                  | SOME n => n
            in
                case !ebr of
                    EvFree (condeps, moddeps, id) =>
                        let fun concatWith s l =
                                case l of
                                    [] => ""
                                  | [x] => x
                                  | x::xs => x ^ s ^ concatWith s xs
                            val cdl =
                            ConVarSet.foldl
                                (fn (cx, l) => Variable.toString cx :: l)
                                []
                                condeps
                            val mdl =
                            ModVarSet.foldl
                                (fn (mx, l) => Variable.toString mx :: l)
                                []
                                moddeps
                            val cds = concatWith "," cdl
                            val mds = concatWith "," mdl
                        in
                            "E" ^ Int.toString n ^ "["
                            ^ "{" ^ cds ^ "};"
                            ^ "{" ^ mds ^ "}"
                            ^ "]"
                        end
                  | _ => raise Fail "impossible"
            end
    end


    (*** kinds ***)

    fun atomic_kind_to_layout knd =
        case knd of
            KndType => $"Type"
          | KndProd ks =>
                &[ $"*",
                   Layout.sequence ("[", "]", ",") (map kind_to_layout ks) ]
          | KndSing c => &[$"S(", con_to_layout c, $")"]
          | _ => Layout.paren (kind_to_layout knd)

    and kind_to_layout knd =
        case knd of
            KndPi (cx, k, k') =>
                %[ &[$"Pi ", var cx, $":", atomic_kind_to_layout k, $"."],
                   kind_to_layout k']
          | _ => atomic_kind_to_layout knd


    (*** constructors ***)

    and atomic_con_to_layout con =
        let fun labcon (lab, con) = %[ &[label lab, $" :"],
                                       con_to_layout con ]
        in
            case con of
                ConVar cx => var cx
              | TyLit lt =>
                    (case lt of
                        Literal.LitTyInt => $"int"
                      | Literal.LitTyChar => $"char"
                      | Literal.LitTyString => $"string")
              | TyTagged => $"tagged"
              | TyRecord labcons =>
                    &[$"*",
                      Layout.sequence ("[", "]", ",") (map labcon labcons)]
              | TySum labcons =>
                    &[$"+",
                      Layout.sequence ("[", "]", ",") (map labcon labcons)]
              | ConEvar ebr =>
                    (case !ebr of
                        EvFree _ => $(ebr_to_string ebr)
                      | EvBound _ => raise Fail "unnormalized evar!")
              | ConProd cs =>
                    Layout.sequence ("<", ">", ",")
                        (map con_to_layout cs)
              | ConModProj (m, lab, labs) =>
                    &[module_to_layout m, $".",
                        Layout.sequence ("[", "]", ",") (map label labs), label lab]
              | _ => Layout.paren (con_to_layout con)
        end

    and app_con_to_layout con =
        case con of
            TyRef conarg =>
                &[$"ref ", atomic_con_to_layout conarg]
          | TyTag conarg =>
                &[$"tag ", atomic_con_to_layout conarg]
          | ConApp (c1, c2) =>
                %[app_con_to_layout c1,
                  atomic_con_to_layout c2]
          | ConProj (i, c) =>
                %[ &[$"#", $(Int.toString i)],
                   atomic_con_to_layout c ]
          | _ => atomic_con_to_layout con

    and arrow_con_to_layout con =
        case con of
            TyArrow (con1, con2) =>
                %[app_con_to_layout con1,
                  &[$"-> ", arrow_con_to_layout con2]]
          | _ => app_con_to_layout con

    and con_to_layout con =
        (* first, eliminate bound evars *)
        let val con = evar_normalize_con con
        in
            case con of
                TyForall (cx, k, c) =>
                    %[ &[$"All ", var cx, $":", atomic_kind_to_layout k, $"."],
                       con_to_layout c ]
              | ConAbs (cx, k, c) =>
                    %[ &[$"\\ ", var cx, $":", atomic_kind_to_layout k, $"."],
                       con_to_layout c ]
              | ConMu (cx, k, c) =>
                    %[ &[$"mu ", var cx, $":", atomic_kind_to_layout k, $"."],
                       con_to_layout c ]
              | _ => arrow_con_to_layout con
        end


    (*** expressions ***)

    and atomic_exp_to_layout exp =
        case exp of
            ExVar x => var x
          | ExLit lit =>
                (case lit of
                    Literal.LitInt i => $(Int.toString i)
                  | Literal.LitChar c => $("'" ^ Char.toString c ^ "'")
                  | Literal.LitString s => $("\"" ^ s ^ "\""))
          | ExRecord labes =>
                let fun labexp (lab, exp) =
                        %[ &[label lab, $" ="], exp_to_layout exp ]
                in
                    %[Layout.sequence ("{", "}", ",") (map labexp labes)]
                end
          | ExFix fbnds =>
                let fun fbnds_to_layout fbnds =
                    map (fn (f, x, c1, c2, e) =>
                        %[ &[var f, $" (", var x, $" : ",
                                       con_to_layout c1, $")"],
                           Layout.indent 2 (&[$": ", con_to_layout c2, $" |->"]),
                           Layout.indent 4 (exp_to_layout e) ])
                        fbnds
                in
                    %[$"fix",
                      Layout.indent 4 (%(fbnds_to_layout fbnds)),
                      $"end"]
                end
          | ExModProj (m, lab, labs) =>
                &[module_to_layout m, $".", Layout.sequence("[", "]", ",") (map label labs), label lab]
          | ExLet (x, e1, e2) =>
                Layout.align
                [ %[ &[$"let ", var x, $" ="],
                     Layout.indent 4 (exp_to_layout e1) ],
                  $"in",
                  Layout.indent 4 (exp_to_layout e2),
                  $"end" ]
          | ExModLet (mx, m, e, c) =>
                Layout.align
                [ %[ &[$"let ", var mx, $" ="],
                     Layout.indent 4 (module_to_layout m) ],
                  $"in",
                  %[ &[Layout.indent 4 (exp_to_layout e), $" :"],
                     con_to_layout c ],
                  $"end" ]
          | _ => Layout.paren (exp_to_layout exp)

    and app_exp_to_layout exp =
        let fun bracket_con c = &[$"[", con_to_layout c, $"]"]
        in
            case exp of
                ExApp (e1, e2) => %[app_exp_to_layout e1,
                                    atomic_exp_to_layout e2]
              | ExConApp (e, c) => %[app_exp_to_layout e,
                                     bracket_con c]
              | ExRecordProj (lab, e, labs) => %[ &[$"#", label lab],
                    Layout.sequence("[", "]", ",") (map label labs), exp_to_layout e]
              | ExRaise (c, e) =>
                    %[ $"raise", Layout.tuple [con_to_layout c,
                                                exp_to_layout e] ]
              | ExRef e =>
                    %[ $"ref", atomic_exp_to_layout e ]
              | ExGet e =>
                    %[ $"get", atomic_exp_to_layout e ]
              | ExSet (e1, e2) =>
                    %[ $"set", Layout.tuple [exp_to_layout e1,
                                              exp_to_layout e2] ]
              | ExRoll (c, e) =>
                    %[ $"roll", bracket_con c, atomic_exp_to_layout e ]
              | ExUnroll e =>
                    %[ $"unroll", atomic_exp_to_layout e ]
              | ExSumInj (c, lab, e, labs) =>
                    %[ $"inj", bracket_con c,
                       Layout.sequence("[", "]", ",") (map label labs),
                       label lab, atomic_exp_to_layout e ]
              | ExTagInj (e1, e2) =>
                    %[ $"tag", Layout.tuple [exp_to_layout e1,
                                             exp_to_layout e2] ]
              | ExNewTag c =>
                    %[ $"newtag", bracket_con c ]
              | ExUnOp (oper, e) =>
                    %[ $(Literal.unop_to_string oper), atomic_exp_to_layout e]
              | _ => atomic_exp_to_layout exp
        end
    
    and times_exp_to_layout exp =
        case exp of
            ExBinOp (e1, Literal.OpTimes, e2) =>
                %[times_exp_to_layout e1,
                  &[$"* ", app_exp_to_layout e2]]
          | _ => app_exp_to_layout exp
    
    and plus_minus_exp_to_layout exp =
        case exp of
            ExBinOp (e1, Literal.OpPlus, e2) =>
                %[plus_minus_exp_to_layout e1,
                  &[$"+ ", times_exp_to_layout e2]]
          | ExBinOp (e1, Literal.OpMinus, e2) =>
                %[plus_minus_exp_to_layout e1,
                  &[$"- ", times_exp_to_layout e2]]
          | ExBinOp (e1, Literal.OpConcat, e2) =>
                %[plus_minus_exp_to_layout e1,
                  &[$"^ ", times_exp_to_layout e2]]
          | _ => times_exp_to_layout exp

    and equals_exp_to_layout exp =
        case exp of
            ExBinOp (e1, Literal.OpEquals, e2) =>
                %[plus_minus_exp_to_layout e1,
                  &[$"= ", plus_minus_exp_to_layout e2]]
          | _ => plus_minus_exp_to_layout exp

    and exp_to_layout exp =
        case exp of
            ExIfTagOf (e1, e2, e3, e4) =>
                Layout.align
                [ &[$"iftagof ", exp_to_layout e1, $" is ",
                                 exp_to_layout e2, $" then"],
                  Layout.indent 4 (exp_to_layout e3),
                  $"else",
                  Layout.indent 4 (exp_to_layout e4) ]
          | ExCase (c, e, labes) =>
                let fun labexp (lab, exp) =
                        %[ &[label lab, $" |->"], exp_to_layout exp ]
                in
                    Layout.align
                    [ &[$"case ", exp_to_layout e, $" of"],
                      Layout.indent 4 (Layout.sequence ("{", "}", ",")
                                        (map labexp labes)) ]
                end
          | ExConAbs (cx, k, e) =>
                %[ &[$"/\\ ", var cx, $":", atomic_kind_to_layout k, $"."],
                   exp_to_layout e]
          | ExHandle (e1, e2) =>
                %[ $"handle",
                   Layout.indent 4 (exp_to_layout e1),
                   $"with",
                   Layout.indent 4 (exp_to_layout e2) ]
          | _ => equals_exp_to_layout exp


    (*** modules ***)

    and atomic_module_to_layout module =
        case module of
            ModVar mx => var mx
          | ModStruct sbnds =>
                &[ $"[", sbnds_to_layout sbnds, $"]" ]
          | ModModProj (m, lab, labs) =>
                &[module_to_layout m, $".",
                  Layout.sequence("[", "]", ",") (map label labs), label lab]
          | ModLet (mx, m, m', sg') =>
                Layout.align
                [ %[ &[$"let ", var mx, $" ="],
                     Layout.indent 4 (module_to_layout m) ],
                  $"in",
                  %[ &[Layout.indent 4 (module_to_layout m'), $" :"],
                     signat_to_layout sg' ],
                  $"end" ]
          | _ => Layout.paren (module_to_layout module)

    and app_module_to_layout module =
        case module of
            ModApp (m1, m2) =>
                %[app_module_to_layout m1, atomic_module_to_layout m2]
          | _ => atomic_module_to_layout module

    and module_to_layout module =
        case module of
            ModAbs (mx, sg, m) =>
                %[ &[$"\\ ", var mx, $":", signat_to_layout sg, $"."],
                   module_to_layout m]
          | ModConstrain (m, sg) =>
                %[module_to_layout m,
                  &[$":> ", signat_to_layout sg]]
          | _ => app_module_to_layout module
        
    and sbnds_to_layout sbnds =
        %(Layout.separateRight (map sbnd_to_layout sbnds, ","))

    and sbnd_to_layout sbnd =
        case sbnd of
            (lab, BndExp (ex, e)) =>
                (*
                &[label lab, $" ",
                  %[ &[$"|> ", var ex],
                     Layout.indent 1 (&[$"= ", exp_to_layout e]) ] ]
                *)
                %[ &[label lab, $" |> ", var ex, $" ="],
                   Layout.indent 4 (exp_to_layout e) ]
          | (lab, BndCon (cx, c)) =>
                %[ &[label lab, $" |> ", var cx, $" ="],
                   Layout.indent 4 (con_to_layout c) ]
          | (lab, BndMod (mx, m)) =>
                %[ &[label lab, $" |> ", var mx, $" ="],
                   Layout.indent 4 (module_to_layout m) ]

    and signat_to_layout signat =
        case signat of
            SigStruct sdecs =>
                &[ $"[", sdecs_to_layout sdecs, $"]" ]
          | SigPi (mx, sg, sg') =>
                %[ &[$"Pi ", var mx, $":", signat_to_layout sg, $"."],
                   signat_to_layout sg']

    and sdecs_to_layout sdecs =
        %(Layout.separateRight (map sdec_to_layout sdecs, ","))
    
    and sdec_to_layout sdec =
        case sdec of
            (lab, DecExpCon (ex, c)) =>
                %[ &[label lab, $" |> ", var ex, $" :"],
                   Layout.indent 4 (con_to_layout c) ]
          | (lab, DecConKind (cx, k)) =>
                %[ &[label lab, $" |> ", var cx, $" :"],
                   Layout.indent 4 (kind_to_layout k) ]
          | (lab, DecModSig (mx, sg)) =>
                %[ &[label lab, $" |> ", var mx, $" :"],
                   Layout.indent 4 (signat_to_layout sg) ]

    (* for debugging, printing contexts *)
    fun dec_to_layout dec =
        case dec of
            DecExpCon (ex, c) =>
                %[ &[var ex, $" :"], Layout.indent 4 (con_to_layout c) ]
          | DecConKind (cx, k) =>
                %[ &[var cx, $" :"], Layout.indent 4 (kind_to_layout k) ]
          | DecModSig (mx, sg) =>
                %[ &[var mx, $" :"], Layout.indent 4 (signat_to_layout sg) ]

    fun ctx_to_layout ctx =
        Layout.tuple (map dec_to_layout (List.rev ctx))

    val program_to_layout = sbnds_to_layout
  end
