structure IL2Static :> IL2STATIC =
  struct
    open IL2
    structure Ctx = ContextFn(struct 
                                type kind = kind 
                                type con = con
                              end)
    type context = Ctx.context

    exception Equiv
    exception Occurs
    exception Subkind
    exception Check of string
    exception Error of string

    fun check b f = if b then () else f ()

    fun nth (l, n) = SOME (List.nth (l,n)) handle Subscript => NONE


    (*** static semantics ***)

    (* precondition: con is a path *)
    fun extract_natural_kind ctx con =
        case con of
            ConVar cx =>
                (case Ctx.lookup_convar ctx cx of
                    NONE => raise Check
                            ("extract_natural_kind: unbound convar, "
                            ^ Variable.toString cx)
                  | SOME k => k)
          | ConApp (p, c) =>
                (case extract_natural_kind ctx p of
                    KndPi (cx, k, l) => subst_con_kind c cx l
                  | _ => raise Error
                         "extract_natural_kind: function not of Pi kind")
          | ConProj1 p =>
                (case extract_natural_kind ctx p of
                    KndSigma (cx, k1, k2) => k1
                  | _ => raise Error
                         "extract_natural_kind: projection from non-pair")
          | ConProj2 p =>
                (case extract_natural_kind ctx p of
                    KndSigma (cx, k1, k2) => subst_con_kind (ConProj1 p) cx k2
                  | _ => raise Error
                         "extract_natural_kind: projection from non-pair")
          | ConMu (cx, k, c) => k
          | _ => raise Error "extract_natural_kind: applied to a non-path"
          (* XXX ConUnit? *)

    fun whnf ctx con =
        let fun path_reduce ctx conpath =
                case extract_natural_kind ctx conpath of
                    KndSing c => SOME c
                  | _ => NONE
            fun reduce ctx con =
                case con of
                  (* beta reductions *)
                    ConApp (ConAbs (cx, k, c1), c2) =>
                        SOME (subst_con_con c2 cx c1)
                  | ConProj1 (ConPair (c1, _)) => SOME c1
                  | ConProj2 (ConPair (_, c2)) => SOME c2
                  (* congruences *)
                  | ConApp (c1, c2) =>
                        (case reduce ctx c1 of
                            SOME c1' => SOME (ConApp (c1', c2))
                          | NONE => path_reduce ctx con)
                  | ConProj1 c =>
                        (case reduce ctx c of
                            SOME c' => SOME (ConProj1 c')
                          | NONE => path_reduce ctx con)
                  | ConProj2 c =>
                        (case reduce ctx c of
                            SOME c' => SOME (ConProj2 c')
                          | NONE => path_reduce ctx con)
                  (* definition resolution *)
                  | ConVar _ => path_reduce ctx con
                  | _ => NONE
        in
            case reduce ctx con of
                SOME con' => whnf ctx con'
              | NONE => con
        end

    fun subkind ctx knd1 knd2 =
        case (knd1, knd2) of
            (KndType, KndType) => ()
          | (KndSing c, KndType) => ()
          | (KndSing c1, KndSing c2) =>
                let val () = equiv_con ctx c1 c2 KndType
                             handle Equiv => raise Subkind
                in
                    ()
                end
          | (KndPi (cx1, k1, l1), KndPi (cx2, k2, l2)) =>
                let val () = subkind ctx k2 k1
                    val (cx, (l1, l2)) = IL2.alpha_vary2 (cx1, l1) (cx2, l2)
                                                     replace_con_kind
                    val ctx = Ctx.extend ctx (Ctx.DecConKind (cx, k2))
                    val () = subkind ctx l1 l2
                in
                    ()
                end
          | (KndUnit, KndUnit) => ()
          | (KndSigma (cx1, k1, l1), KndSigma (cx2, k2, l2)) =>
                let val () = subkind ctx k1 k2
                    val (cx, (l1, l2)) = IL2.alpha_vary2 (cx1, l1) (cx2, l2)
                                                     replace_con_kind
                    val ctx = Ctx.extend ctx (Ctx.DecConKind (cx, k1))
                    val () = subkind ctx l1 l2
                in
                    ()
                end
          | _ => raise Subkind

    and equiv_con ctx con1 con2 knd =
        case knd of
            KndType =>
                let val k = equiv_con_path ctx (whnf ctx con1) (whnf ctx con2)
                    (* is this kind check necessary?  maybe for the sake of
                       belt and suspenders... *)
                    val () = subkind ctx k KndType
                             handle Equiv => raise Error
                             ("mismatch between equiv_con input (KndType) "
                             ^ " and equiv_con_path output")
                in
                    ()
                end
          | KndSing _ => ()
          | KndUnit => ()
          | KndSigma (cx, k1, k2) =>
                let val () = equiv_con ctx (ConProj1 con1) (ConProj1 con2) k1
                    val () = equiv_con ctx (ConProj2 con1) (ConProj2 con2)
                                        (subst_con_kind (ConProj1 con1) cx k2)
                in
                    ()
                end
          | KndPi (x, k, l) =>
                let val (x, l) = alpha_vary (x, l) replace_con_kind
                in
                    equiv_con (Ctx.extend ctx (Ctx.DecConKind (x, k)))
                              (ConApp (con1, ConVar x))
                              (ConApp (con2, ConVar x))
                              l
                end

    and equiv_con_path ctx con1 con2 =
        case (con1, con2) of
            (ConVar cx1, ConVar cx2) =>
                if Variable.eq (cx1, cx2) then
                    case Ctx.lookup_convar ctx cx1 of
                        NONE => raise Error ("comparing unbound variable, "
                                            ^ Variable.toString cx1)
                      | SOME k => k
                else
                    raise Equiv
          | (ConApp (cp1, c1), ConApp (cp2, c2)) =>
                let val k = equiv_con_path ctx cp1 cp2
                in
                    case k of
                        KndPi (x, k1, k2) =>
                            let val () = equiv_con ctx c1 c2 k1
                            in
                                subst_con_kind c1 x k2
                            end
                      | _ => raise Check
                             "equiv_con_path: application of non-function"
                end
          | (ConProj1 cp1, ConProj1 cp2) =>
                let val k = equiv_con_path ctx cp1 cp2
                in
                    case k of
                        KndSigma (cx, k1, k2) => k1
                      | _ => raise Check
                             "equiv_con_path: ConProj1 from a non-pair"
                end
          | (ConProj2 cp1, ConProj2 cp2) =>
                let val k = equiv_con_path ctx cp1 cp2
                in
                    case k of
                        KndSigma (cx, k1, k2) =>
                            subst_con_kind (ConProj1 cp1) cx k2
                      | _ => raise Check
                             "equiv_con_path: ConProj1 from a non-pair"
                end
          | (ConMu (cx1, k1, c1), ConMu (cx2, k2, c2)) =>
                let val () = equiv_kind ctx k1 k2
                    val (cx, (c1, c2)) = alpha_vary2 (cx1, c1) (cx2, c2)
                                                     replace_con_con
                    val () = equiv_con (Ctx.extend ctx (Ctx.DecConKind (cx,k1)))
                                       c1 c2 k1
                in
                    k1
                end

          | (TyLit lt1, TyLit lt2) =>
                let val () = check (Literal.lit_ty_eq (lt1, lt2))
                                   (fn () => raise Equiv)
                in
                    KndType
                end
          | (TyArrow (c1dom, c1cod), TyArrow (c2dom, c2cod)) =>
                let val () = equiv_con ctx c1dom c2dom KndType
                    val () = equiv_con ctx c1cod c2cod KndType
                in
                    KndType
                end
          | (TyRef c1, TyRef c2) =>
                let val () = equiv_con ctx c1 c2 KndType
                in
                    KndType
                end
          | (TyTagged, TyTagged) => KndType
          | (TyTag c1, TyTag c2) =>
                let val () = equiv_con ctx c1 c2 KndType
                in
                    KndType
                end
          | (TySum cs1, TySum cs2) =>
                let val () = check (List.length cs1 = List.length cs2)
                                   (fn () => raise Equiv)
                    val () = ListPair.app
                                (fn (c1, c2) => equiv_con ctx c1 c2 KndType)
                                (cs1, cs2)
                in
                    KndType
                end
          | (TyProd cs1, TyProd cs2) =>
                let val () = check (List.length cs1 = List.length cs2)
                                   (fn () => raise Equiv)
                    val () = ListPair.app
                                (fn (c1, c2) => equiv_con ctx c1 c2 KndType)
                                (cs1, cs2)
                in
                    KndType
                end
          | (TyForall (cx1, k1, c1), TyForall (cx2, k2, c2)) =>
                let val () = equiv_kind ctx k1 k2
                    val (cx, (c1, c2)) = alpha_vary2 (cx1, c1) (cx2, c2)
                                                     replace_con_con
                    val () = equiv_con (Ctx.extend ctx (Ctx.DecConKind (cx,k1)))
                                       c1 c2 KndType
                 in
                     KndType
                 end

          | (ConAbs _, ConAbs _) => raise Error "equiv_con_path on non-path"
          | (ConUnit, ConUnit) => raise Error "equiv_con_path on non-path"
          | (ConPair _, ConPair _) => raise Error "equiv_con_path on non-path"
          | _ => raise Equiv


    and check_equiv_con ctx con1 con2 knd s =
        equiv_con ctx con1 con2 knd
        handle Equiv => raise Check s

    and equiv_kind ctx knd1 knd2 =
        case (knd1, knd2) of
            (KndType, KndType) => ()
          | (KndPi (cx1, k1, l1), KndPi (cx2, k2, l2)) =>
                let val () = equiv_kind ctx k1 k2
                    val (cx, (l1, l2)) = alpha_vary2 (cx1, l1) (cx2, l2)
                                                     replace_con_kind
                    val ctx = Ctx.extend ctx (Ctx.DecConKind (cx, k1))
                    val () = equiv_kind ctx l1 l2
                in
                    ()
                end
          | (KndUnit, KndUnit) => ()
          | (KndSigma (cx1, k1, l1), KndSigma (cx2, k2, l2)) =>
                let val () = equiv_kind ctx k1 k2
                    val (cx, (l1, l2)) = alpha_vary2 (cx1, l1) (cx2, l2)
                                                     replace_con_kind
                    val ctx = Ctx.extend ctx (Ctx.DecConKind (cx, k1))
                    val () = equiv_kind ctx l1 l2
                in
                    ()
                end
          | (KndSing c1, KndSing c2) => equiv_con ctx c1 c2 KndType
          | _ => raise Equiv

    (* XXX error-handling could probably be better on these check_* functions,
       but since they're mostly belt-and-suspenders checks, i suppose it's okay
       to keep them as they are for now... *)

    and check_kind ctx knd s =
        case knd of
            KndType => ()
          | KndPi (cx, k, l) =>
                let val () = check_kind ctx k s
                    val (cx, l) = alpha_vary (cx, l) replace_con_kind
                    val ctx = Ctx.extend ctx (Ctx.DecConKind (cx, k))
                    val () = check_kind ctx l s
                in
                    ()
                end
          | KndUnit => ()
          | KndSigma (cx, k, l) =>
                let val () = check_kind ctx k s
                    val (cx, l) = alpha_vary (cx, l) replace_con_kind
                    val ctx = Ctx.extend ctx (Ctx.DecConKind (cx, k))
                    val () = check_kind ctx l s
                in
                    ()
                end
          | KndSing c => check_con_kind ctx c KndType s

    and synth_con_kind ctx con =
        case con of
            ConVar cx => 
                (case Ctx.lookup_convar ctx cx of
                    SOME k => labelledKndSing con k
                  | NONE => raise Check ("synth_con_kind: "
                            ^ "unbound convar, " ^ Variable.toString cx))
          | ConAbs (cx, k, c) =>
                let val () = check_kind ctx k ("synth_con_kind: "
                                              ^ "ConAbs over invalid kind")
                    val (cx, c) = alpha_vary (cx, c) replace_con_con
                    val ctx = Ctx.extend ctx (Ctx.DecConKind (cx, k))
                    val k' = synth_con_kind ctx c
                in
                    KndPi (cx, k, k')
                end
          | ConApp (c1, c2) =>
                let val k1 = synth_con_kind ctx c1
                in
                    case k1 of
                        KndPi (cx, k, k') =>
                            let val () = check_con_kind ctx c2 k
                                         ("synth_con_kind: "
                                         ^ "argument in ConApp has wrong kind")
                            in
                                subst_con_kind c2 cx k'
                            end
                      | _ => raise Check
                             "synth_con_kind: function in ConApp has wrong kind"
                end
          | ConUnit => KndUnit
          | ConPair (c1, c2) => 
                kndPair (synth_con_kind ctx c1, synth_con_kind ctx c2)
          | ConProj1 c =>
                let val k = synth_con_kind ctx c
                in
                    case k of
                        KndSigma (cx, k1, k2) => k1
                      | _ => raise Check
                             "synth_con_kind: ConProj1 from a non-pair"
                end
          | ConProj2 c =>
                let val k = synth_con_kind ctx c
                in
                    case k of
                        KndSigma (cx, k1, k2) =>
                            subst_con_kind (ConProj1 c) cx k2
                      | _ => raise Check
                             "synth_con_kind: ConProj2 from a non-pair"
                end
          | ConMu (cx, k, c) =>
                let fun is_Tn k =
                        case k of
                            KndSigma (cx, k1, k2) =>
                                (let val () = equiv_kind ctx k1 KndType
                                 in
                                     is_Tn k2
                                 end
                                 handle Equiv => false)
                          | KndUnit => true
                          | _ => false
                    val () = check_kind ctx k ("synth_con-kind: "
                                              ^ "ConMu over invalid kind")
                    val () = check (is_Tn k) (fn () => raise Check
                                     "synth_con_kind: ConMu over non-T^n kind")
                    val (cx, c) = alpha_vary (cx, c) replace_con_con
                    val ctx = Ctx.extend ctx (Ctx.DecConKind (cx, k))
                    val () = check_con_kind ctx c k
                             "synth_con_kind: ConMu body has wrong kind"
                in
                    labelledKndSing con k
                end
          | TyLit _ => KndSing con
          | TyArrow (c1, c2) =>
                let val () = check_con_kind ctx c1 KndType 
                            "synth_con_kind: TyArrow with non-type component"
                    val () = check_con_kind ctx c2 KndType
                            "synth_con_kind: TyArrow with non-type component"
                in
                    KndSing con
                end
          | TyRef c =>
                let val () = check_con_kind ctx c KndType
                             "synth_con_kind: TyRef with non-type argument"
                in
                    KndSing con
                end
          | TyTagged => KndSing con
          | TyTag c =>
                let val () = check_con_kind ctx c KndType
                             "synth_con_kind: TyTag with non-type argument"
                in
                    KndSing con
                end
          | TySum cs =>
                let val () = List.app
                                (fn c => check_con_kind ctx c KndType
                                          ("synth_con_kind: TySum "
                                          ^ "with non-type component"))
                                cs
                in
                    KndSing con
                end
          | TyProd cs =>
                let val () = List.app
                                (fn c => check_con_kind ctx c KndType
                                          ("synth_con_kind: TyProd "
                                          ^ "with non-type component"))
                                cs
                in
                    KndSing con
                end
          | TyForall (cx, k, c) =>
                let val () = check_kind ctx k ("synth_con_kind: "
                                              ^ "TyForall over invalid kind")
                    val (cx, c) = alpha_vary (cx, c) replace_con_con
                    val ctx = Ctx.extend ctx (Ctx.DecConKind (cx, k))
                    val () = check_con_kind ctx c KndType
                             "synth_con_kind: TyForall with non-type body"
                in
                    KndSing con
                end

    and check_con_kind ctx con knd s =
        let val knd' = synth_con_kind ctx con
        in
            subkind ctx knd' knd
            handle Subkind => raise Check s
        end

    (* expansion for recursive kinds, a la Dreyer *)
    fun expand ctx recpath =
        let (* decomposes a path into its elimination context and its root
               if [ decompose_path path ] = (elim_ctx, root),
               then [ elim_ctx root ] = path *)
            fun decompose_path path =
                case path of
                    ConProj1 cp =>
                        let val (elim_ctx, root) = decompose_path cp
                        in
                            (fn c => ConProj1 (elim_ctx c), root)
                        end
                  | ConProj2 cp =>
                        let val (elim_ctx, root) = decompose_path cp
                        in
                            (fn c => ConProj2 (elim_ctx c), root)
                        end
                  (*
                  (* don't handle apps -- only permit uniform datatypes *)
                  | ConApp (cp, c_arg) =>
                        let val (elim_ctx, root) = decompose_path cp
                        in
                            (fn c => ConApp (elim_ctx c, c_arg), root)
                        end
                  *)
                  | _ => (fn c => c, path)
            val () = check_con_kind ctx recpath KndType
                     "expand: recursive path is not a type"
            val () = equiv_kind ctx (extract_natural_kind ctx recpath) KndType
                     handle Equiv => raise Check
                     "expand: recursive path's natural kind not KndType"
            val (elim_ctx, c_mu) = decompose_path recpath
        in
            case c_mu of
                ConMu (cx, k, c) => elim_ctx (subst_con_con c_mu cx c)
              | _ => raise Check "expand: recursive path not rooted at ConMu"
        end

    and synth_exp_con ctx exp =
        case exp of
            ExVar x => 
                (case Ctx.lookup_expvar ctx x of
                    SOME ty => ty
                  | NONE => raise Check
                              ("unbound expvar, " ^ Variable.toString x))
          | ExLit l => TyLit (Literal.lit_ty l)
          | ExUnOp(oper, e) =>
                let val (cin, cout) = Literal.unop_tys oper TyLit tyBool tyUnit
                    val () = check_exp_con ctx e cin
                             "synth_exp_con: ExUnOp argument incorrect"
                in
                  cout
                end
          | ExBinOp(e1, oper, e2) =>
                let val (cin1, cin2, cout) =
                          Literal.binop_tys oper TyLit tyBool tyUnit
                    val () = check_exp_con ctx e1 cin1
                             "synth_exp_con: ExBinOp first argument incorrect"
                    val () = check_exp_con ctx e2 cin2
                             "synth_exp_con: ExBinOp second argument incorrect"
                in
                    cout
                end
          | ExLam(x, c, e) =>
                let
                  val (x, e) = alpha_vary (x, e) replace_exp_exp
                  val () = check_con_kind ctx c KndType
                        ("synth_exp_con: Constructor at lambda " ^
                         "binding not valid at kind Type.")
                  val ctx = Ctx.extend ctx (Ctx.DecExpCon(x, c))
                  val c' = synth_exp_con ctx e
                in
                  TyArrow(c, c')
                end
          | ExFix fbnds =>
                let (* every function body is checked in an enriched scope
                       including bindings for all the functions *)
                       (* XXX should technically be doing some massive
                          alpha-varying, but this should work since no
                          classifier can depend on expression variables *)
                    val ctx0 = ctx (* save original context *)
                    val ctx = List.foldl
                                (fn ((f, _, c, c', _), ctx) =>
                                      Ctx.extend ctx
                                      (Ctx.DecExpCon (f, TyArrow (c, c'))))
                                ctx
                                fbnds
                    (* each individual function body is checked in a scope
                       enriched with the type of its formal parameter *)
                    fun check_fbnd ctx (f, x, c, cbody, ebody) =
                        let val ctx = Ctx.extend ctx (Ctx.DecExpCon (x, c))
                            val () = check_con_kind ctx0 c KndType
                                     "synth_exp_con: ExFix domain not a type"
                            val () = check_con_kind ctx0 cbody KndType
                                     "synth_exp_con: ExFix co-domain not a type"
                            val () = check_exp_con ctx ebody cbody
                                   "synth_exp_con: function body has wrong type"
                        in
                            ()
                        end
                    val () = List.app (check_fbnd ctx) fbnds
                in
                    (* the resulting type is a product of functions *)
                    TyProd (List.map (fn (_, _, c, c', _) => TyArrow (c, c'))
                                     fbnds)
                end
          | ExApp (e1, e2) =>
                let val c1 = synth_exp_con ctx e1
                in
                    case whnf ctx c1 of
                        TyArrow (c1dom, c1cod) =>
                            (* XXX use check_exp_con instead? *)
                            let val c2 = synth_exp_con ctx e2
                                val () = check_equiv_con ctx c1dom c2 KndType
                                         ("synth_exp_con: function argument "
                                         ^ "doesn't match function domain")
                            in
                                c1cod
                            end
                      | _ => raise Check
                             "synth_exp_con: application of a non-function"
                end
          | ExHandle (e1, e2) =>
                (* XXX could be simpler, at a risk of worse error reporting,
                   by just using check_exp_con on e2 *)
                let val c1 = synth_exp_con ctx e1
                    val c2 = synth_exp_con ctx e2
                in
                    case whnf ctx c2 of
                        TyArrow (tytagged, c1') =>
                            (case whnf ctx tytagged of
                                TyTagged =>
                                    let val () = check_equiv_con ctx c1 c1'
                                                 KndType
                                                 ("exception handler has "
                                                 ^ "wrong return type")
                                    in
                                        c1
                                    end
                              | _ => raise Check ("synth_exp_con: exception "
                                     ^ "handler domain type not TyTagged"))
                      | _ => raise Check ("synth_exp_con: "
                             ^ "exception handler not of arrow type")
                end
          | ExRaise (c, e) =>
                let val () = check_exp_con ctx e TyTagged
                             "synth_exp_con: raise arg not of type TyTagged"
                    val () = check_con_kind ctx c KndType
                             "synth_exp_con: con arg to raise not a type"
                in
                    c
                end
          | ExRef e => TyRef (synth_exp_con ctx e)
          | ExGet e =>
                let val c = synth_exp_con ctx e
                in
                    case whnf ctx c of
                        TyRef c' => c'
                      | _ => raise Check
                             "synth_exp_con: ExGet applied to a non-ref"
                end
          | ExSet (e1, e2) =>
                let val c1 = synth_exp_con ctx e1
                in
                    case whnf ctx c1 of
                        TyRef c2 =>
                            let val () = check_exp_con ctx e2 c2
                                         ("synth_exp_con: ExSet arg has "
                                         ^ "wrong type")
                            in
                                tyUnit
                            end
                      | _ => raise Check
                             "synth_exp_con: ExSet applied to a non-ref"
                end
          (* see Dreyer's dissertation for the Roll/Unroll rules *)
          | ExRoll (c, e) =>
                let val () = check_con_kind ctx c KndType
                             "synth_exp_con: ExRoll at non-type"
                    val recpath = whnf ctx c
                    val c_expanded = expand ctx recpath
                    val () = check_exp_con ctx e c_expanded
                             "synth_exp_con: ExRoll arg has wrong type"
                in
                    recpath
                end
          | ExUnroll e =>
                let val c = synth_exp_con ctx e
                    val recpath = whnf ctx c
                    val c_expanded = expand ctx recpath
                in
                    c_expanded
                end
          | ExProd es => TyProd (map (synth_exp_con ctx) es)
          | ExProj (i, e) =>
                let val c = synth_exp_con ctx e
                in
                    case whnf ctx c of
                        TyProd tys =>
                            (case nth (tys, i-1) of
                                NONE => raise Check ("synth_exp_con: "
                                        ^ "ExProj subscript out of bounds.")
                              | SOME ty => ty)
                      | _ => raise Check
                             "synth_exp_con: projection from a non-tuple"
                end
          | ExSumInj (c, i, e) =>
                let val () = check_con_kind ctx c KndType
                             "synth_exp_con: ExSumInj at non-type"
                    val c = whnf ctx c
                in
                    case c of
                        TySum tys =>
                            (case nth (tys, i-1) of 
                                SOME ty =>
                                    let val () = check_exp_con ctx e ty
                                                 ("synth_exp_con: ExSumInj "
                                                 ^ "arg has wrong type")
                                    in
                                        c
                                    end
                              | NONE => raise Check ("synth_exp_con: "
                                        ^ "ExSumInj subscript out of bounds"))
                      | _ => raise Check
                             "synth_exp_con: ExSumInj into non-TySum type"
                end
          | ExCase (c_cod, e, es) =>
                let val c_sum = synth_exp_con ctx e
                in
                    case whnf ctx c_sum of
                        TySum tys =>
                            let val () =
                                ListPair.app
                                    (fn (c_dom, e) =>
                                        let val () = check_exp_con ctx e
                                                     (TyArrow (c_dom, c_cod))
                                                     ("synth_exp_con: type "
                                                     ^ "mismatch in ExCase")
                                        in
                                            ()
                                        end)
                                    (tys, es)
                            in
                                c_cod
                            end
                      | _ => raise Check
                             "synth_exp_con: ExCase over non-TySum type"
                end
          | ExTagInj (e1, e2) =>
                let val c1 = synth_exp_con ctx e1
                in
                    case whnf ctx c1 of
                        TyTag c2 =>
                            let val () = check_exp_con ctx e2 c2
                                         ("synth_exp_con: ExTagInj data has "
                                         ^ "incorrect type")
                            in
                                TyTagged
                            end
                      | _ => raise Check "synth_exp_con: ExTagInj into non-tag"
                end
          | ExNewTag c =>
                let val () = check_con_kind ctx c KndType
                             "synth_exp_con: ExNewTag carrying non-type"
                in
                    TyTag c
                end
          | ExIfTagOf (e1, e2, e3, e4) =>
                let val () = check_exp_con ctx e1 TyTagged
                             "synth_exp_con: ExIfTagOf on non-TyTagged"
                    val c_data =
                        case whnf ctx (synth_exp_con ctx e2) of
                            TyTag c => c
                          | _ => raise Check ("synth_exp_con: ExIfTagOf "
                                             ^ "comparison exp not a TyTag")
                    val c_res =
                        case whnf ctx (synth_exp_con ctx e3) of
                            TyArrow (c_data', c) =>
                                let val () = check_equiv_con ctx c_data c_data'
                                             KndType
                                             ("synth_exp_con: ExIfTagOf `then' "
                                             ^ "function has incorrect domain")
                                in
                                    c
                                end
                          | _ => raise Check ("synth_exp_con: ExIfTagOf `then' "
                                             ^ "exp not a function")
                    val () = check_exp_con ctx e4 c_res
                             ("synth_exp_con: ExIfTagOf `else' exp "
                             ^ "has incorrect type")
                in
                    c_res
                end
          | ExConAbs (cx, k, e) =>
                let val () = check_kind ctx k ("synth_exp_con: "
                                              ^ "ExConAbs over invalid kind")
                    val (cx, e) = alpha_vary (cx, e) replace_con_exp
                    val ctx = Ctx.extend ctx (Ctx.DecConKind (cx, k))
                    val c = synth_exp_con ctx e
                in
                    TyForall (cx, k, c)
                end
          | ExConApp (e, c_arg) =>
                let val c_fun = synth_exp_con ctx e
                in
                    case whnf ctx c_fun of
                        TyForall (cx, k, c) =>
                            let val () = check_con_kind ctx c_arg k
                                         ("synth_con_kind: ExConApp arg "
                                         ^ "has wrong kind")
                            in
                                subst_con_con c_arg cx c
                            end
                      | _ => raise Check "con-application of a non-forall"
                end
          | ExLet (ex, e, e') =>
                let val c = synth_exp_con ctx e
                    val (ex, e') = alpha_vary (ex, e') replace_exp_exp
                    val ctx = Ctx.extend ctx (Ctx.DecExpCon (ex, c))
                    val c' = synth_exp_con ctx e'
                in
                    c'
                end

    and check_exp_con ctx exp con s =
        let val con' = synth_exp_con ctx exp
        in
            equiv_con ctx con con' KndType
            handle Equiv => raise Check s
        end

    fun synth_mod_sig ctx (Mod (c, e)) =
        let val cx = Variable.newNamed "_a"
        in
            Sig (cx, synth_con_kind ctx c, synth_exp_con ctx e)
        end

    (* override check_* for export *)
    fun check_kind' ctx k =
        let val () = check_kind ctx k ""
        in
            true
        end
        handle Check "" => false
    val check_kind = check_kind'

    fun check_con_kind' ctx c k =
        let val () = check_con_kind ctx c k ""
        in
            true
        end
        handle Check "" => false
    val check_con_kind = check_con_kind'

    fun check_exp_con' ctx e c =
        let val () = check_exp_con ctx e c ""
        in
            true
        end
        handle Check "" => false
    val check_exp_con = check_exp_con'

    fun check_mod_sig ctx (Mod (c_m, e_m)) (Sig (cx, k_sg, c_sg)) =
        check_con_kind ctx c_m k_sg
        andalso
        check_exp_con ctx e_m (subst_con_con c_m cx c_sg)

    fun check_signat ctx (Sig (cx, k, c)) =
        check_kind ctx k
        andalso
        check_con_kind (Ctx.extend ctx (Ctx.DecConKind (cx, k))) c KndType

    fun equiv_signat ctx (Sig (cx1, k1, c1)) (Sig (cx2, k2, c2)) =
        ((equiv_kind ctx k1 k2; true) handle Equiv => false)
        andalso
        let val (cx, (c1, c2)) = alpha_vary2 (cx1, c1) (cx2, c2) replace_con_con
            val ctx = Ctx.extend ctx (Ctx.DecConKind (cx, k1))
        in
            (equiv_con ctx c1 c2 KndType; true)
            handle Equiv => false
        end

    fun sub_signat ctx (Sig (cx1, k1, c1)) (Sig (cx2, k2, c2)) =
        ((subkind ctx k1 k2; true) handle Subkind => false)
        andalso
        let val (cx, (c1, c2)) = alpha_vary2 (cx1, c1) (cx2, c2) replace_con_con
            val ctx = Ctx.extend ctx (Ctx.DecConKind (cx, k1))
        in
            (equiv_con ctx c1 c2 KndType; true)
            handle Equiv => false
        end
  end
