structure IL3 :> IL3 =
  struct
    type expvar = Variables.exp Variable.var
    type convar = Variables.con Variable.var
    type modvar = Variables.mod Variable.var

    datatype kind =
        KndType
      | KndArrow of kind * kind (* convar bound in kind2 *)
      | KndUnit
      | KndPair of kind * kind (* convar bound in kind2 *)

    datatype con =
        ConVar of convar
      | ConAbs of convar * kind * con
      | ConApp of con * con
      | ConUnit
      | ConPair of con * con
      | ConProj1 of con
      | ConProj2 of con
      | ConMu of convar * kind * con (* convar bound in con, kind = KndType^n *)

      | TyLit of Literal.lit_ty
      | TyArrow of con * con
      | TyRef of con
      | TyTagged
      | TyTag of con
      | TySum of con list
      | TyProd of con list
      | TyForall of convar * kind * con (* convar bound in con *)

    datatype exp =
        ExVar of expvar
      | ExLit of Literal.lit
      | ExUnOp of Literal.unop * exp
      | ExBinOp of exp * Literal.binop * exp
      | ExFix of int * fbnd list (* includes projection *)
      | ExApp of exp * exp
      (* made this bind a variable, for pedagogical purposes *)
      | ExHandle of exp * (expvar * exp) (* expvar bound in exp2 *)
      | ExRaise of con * exp
      | ExRef of exp
      | ExGet of exp
      | ExSet of exp * exp
      | ExRoll of con * exp
      | ExUnroll of exp
      | ExProd of exp list
      | ExProj of int * exp
      | ExSumInj of con * int * exp
      | ExCase of con * exp * exp list
      | ExTagInj of exp * exp
      | ExNewTag of con
      | ExIfTagOf of exp * exp * exp * exp
      | ExConAbs of convar * kind * exp (* convar bound in exp *)
      | ExConApp of exp * con
      | ExLet of expvar * exp * exp (* expvar bound in exp2 *)
    withtype fbnd = (expvar * expvar * con * con * exp)

  (*
   (* should this still be around? *)

    (* degenerate module calculus *)
    datatype module = Mod of con * exp
    datatype signat = Sig of convar * kind * con (* convar bound in con *)
  *)


    (*** derived forms ***)

    fun kndProd ks = foldr KndPair KndUnit ks

    fun conProd cs = foldr ConPair ConUnit cs

    fun conProj (n, c) =
        if n = 1
        then ConProj1 c
        else conProj (n-1, ConProj2 c)

    fun exLam (var, con, con', exp) =
        let val var' = Variable.newNamed "_f"
        in
            ExFix (1, [(var', var, con, con', exp)])
        end


    val tyUnit = TyProd []
    val tyBool = TySum [tyUnit, tyUnit]
    val tyInt = TyLit Literal.LitTyInt
    val tyChar = TyLit Literal.LitTyChar
    val tyString = TyLit Literal.LitTyString

    val exUnit = ExProd []
    fun exBool b = ExSumInj (tyBool, if b then 1 else 2, exUnit)
    fun exInt n = ExLit (Literal.LitInt n)
    fun exChar c = ExLit (Literal.LitChar c)
    fun exString s = ExLit (Literal.LitString s)

    fun exIf con (e1, e2, e3) =
        let val x = Variable.newNamed "_"
        in
            ExCase (con, e1, [exLam (x, tyUnit, con, e2),
                              exLam (x, tyUnit, con, e3)])
        end

    (*** mapping ***)

    (* map_kind no longer needed -- kinds contain only kinds *)

    (* map_con : (convar -> con)
              -> (convar * con -> convar * con)
              -> con -> con *)
    fun map_con fconvar fconvarcon con =
        let fun fcon con = map_con fconvar fconvarcon con
            fun fconvarkindcon (x, k, c) =
                    let val (x, c) = fconvarcon (x, c)
                    in
                        (x, k, c)
                    end
        in
            case con of
                ConVar x => fconvar x
              | ConAbs xkc => ConAbs (fconvarkindcon xkc)
              | ConApp (c1, c2) => ConApp (fcon c1, fcon c2)
              | ConUnit => ConUnit
              | ConPair (c1, c2) => ConPair (fcon c1, fcon c2)
              | ConProj1 c => ConProj1 (fcon c)
              | ConProj2 c => ConProj2 (fcon c)
              | ConMu xkc => ConMu (fconvarkindcon xkc)
              | TyLit lt => con
              | TyArrow (c1, c2) => TyArrow (fcon c1, fcon c2)
              | TyRef c => TyRef (fcon c)
              | TyTagged => con
              | TyTag c => TyTag (fcon c)
              | TySum cons => TySum (map fcon cons)
              | TyProd cons => TyProd (map fcon cons)
              | TyForall xkc => TyForall (fconvarkindcon xkc)
        end

    (* map_exp : (expvar -> exp)
              -> (fbnd list -> fbnd list)
              -> (expvar * exp -> expvar * exp)
              -> (convar * exp -> convar * exp)
              -> (con -> con)
              -> exp -> exp *)
    fun map_exp fexpvar
                ffbnds
                fexpvarexp
                fconvarexp
                fcon
                exp =
        let val fexp = map_exp fexpvar ffbnds fexpvarexp fconvarexp fcon
            fun fconvarkindexp (x, k, e) =
                let val (x, e) = fconvarexp (x, e)
                in
                    (x, k, e)
                end
            fun fexpvarexpexp (x, e1, e2) =
                let val e1 = fexp e1
                    val (x, e2) = fexpvarexp (x, e2)
                in
                    (x, e1, e2)
                end
            fun fexpvarconexp (x, c, e) =
                let val c = fcon c
                    val (x, e) = fexpvarexp (x, e)
                in
                    (x, c, e)
                end
        in
            case exp of
                ExLit _ => exp
              | ExUnOp (oper, e) => ExUnOp (oper, fexp e)
              | ExBinOp (e1, oper, e2) => ExBinOp (fexp e1, oper, fexp e2)
              | ExVar x => fexpvar x
              | ExFix (i, fbnds) => ExFix (i, ffbnds fbnds)
              | ExApp (e1, e2) => ExApp (fexp e1, fexp e2)
              | ExProd exps => ExProd (map fexp exps)
              | ExProj (i, e) => ExProj (i, fexp e)
              | ExHandle (e1, (x, e2)) => ExHandle (fexp e1, fexpvarexp (x, e2))
              | ExRaise (c, e) => ExRaise (fcon c, fexp e)
              | ExRef e => ExRef (fexp e)
              | ExGet e => ExGet (fexp e)
              | ExSet (e1, e2) => ExSet (fexp e1, fexp e2)
              | ExRoll (c, e) => ExRoll (fcon c, fexp e)
              | ExUnroll e => ExUnroll (fexp e)
              | ExSumInj (c, i, e) => ExSumInj (fcon c, i, fexp e)
              | ExCase (c, e, es) => ExCase (fcon c, fexp e, map fexp es)
              | ExTagInj (e1, e2) => ExTagInj (fexp e1, fexp e2)
              | ExNewTag c => ExNewTag (fcon c)
              | ExIfTagOf (e1, e2, e3, e4) => ExIfTagOf (fexp e1, fexp e2,
                                                         fexp e3, fexp e4)
              | ExConAbs xke => ExConAbs (fconvarkindexp xke)
              | ExConApp (e, c) => ExConApp (fexp e, fcon c)
              | ExLet xe1e2 => ExLet (fexpvarexpexp xe1e2)
        end


    (*** replacement of convars ***)

    fun replace_con_con convar convar' con =
        map_con (fn cx =>
                    if Variable.eq (cx, convar) then
                        ConVar convar'
                    else
                        ConVar cx)
                (fn (cx, c) =>
                    if Variable.eq (cx, convar) then
                        (cx, c)
                    else
                        (cx, replace_con_con convar convar' c))
                con

    fun replace_con_exp convar convar' exp =
        map_exp (fn x => ExVar x)
                (fn fbnds =>
                    map (fn (f, x, c1, c2, e) =>
                            (f, x, replace_con_con convar convar' c1,
                                   replace_con_con convar convar' c2,
                                   replace_con_exp convar convar' e))
                        fbnds)
                (fn (ex, e) => (ex, replace_con_exp convar convar' e))
                (fn (cx, e) =>
                    if Variable.eq (cx, convar) then
                        (cx, e)
                    else
                        (cx, replace_con_exp convar convar' e))
                (replace_con_con convar convar')
                exp

    fun replace_exp_exp expvar expvar' exp =
        map_exp (fn x =>
                    if Variable.eq (x, expvar) then
                        ExVar expvar'
                    else
                        ExVar x)
                (fn fbnds =>
                    if List.exists
                        (fn (f, _, _, _, _) => Variable.eq (f, expvar))
                        fbnds
                    then
                        (* expvar is bound in entire phrase, do nothing *)
                        fbnds
                    else
                        map (fn (f, x, c1, c2, e) =>
                                let val e =
                                    if Variable.eq (x, expvar) then
                                        (* expvar is bound in e, do nothing *)
                                        e
                                    else
                                        replace_exp_exp expvar expvar' e
                                in
                                    (f, x, c1, c2, e)
                                end)
                            fbnds)
                (fn (ex, e) =>
                    if Variable.eq (ex, expvar) then
                        (ex, e)
                    else
                        (ex, replace_exp_exp expvar expvar' e))
                (fn (cx, e) => (cx, replace_exp_exp expvar expvar' e))
                (fn c => c) (* expvars do not appear in cons *)
                exp


    fun alpha_vary (c1var, c2obj) replace =
        let val c1var' = Variable.freshen c1var
            val c2obj' = replace c1var c1var' c2obj
        in
            (c1var', c2obj')
        end

    fun alpha_vary2 (varA, objA) (varB, objB) replace =
        let val var' = Variable.freshen varA
            val objA' = replace varA var' objA
            val objB' = replace varB var' objB
        in
            (var', (objA', objB'))
        end

    (* alpha_vary_many : 'a var list * 'b list
                      -> ('a var -> 'a var -> 'b -> 'b)
                      -> 'a var list * 'b list *)
    (* varies many bound variables in many objects. 
       used in subst_exp_exp, for fbnds substitution.*)
    (* not the most efficient -- would be better to do a multiary replace... *)
    fun alpha_vary_many (vars, objs) replace =
        let val vars' = map Variable.freshen vars
            fun subst obj = ListPair.foldl (fn (v, v', ob) => replace v v' ob)
                                           obj
                                           (vars, vars')
            val objs' = map subst objs
        in
            (vars', objs')
        end


    (*** substitution for convars ***)

    fun subst_con_con confor convar incon =
        map_con (fn cx =>
                    if Variable.eq (cx, convar) then
                        (* cx is convar free, substitute for it *)
                        confor
                    else
                        (* not convar -- put it back *)
                        ConVar cx)
                (fn (cx, c) =>
                    if Variable.eq (cx, convar) then
                        (* cx is convar being rebound, leave it alone *)
                        (cx, c)
                    else
                        (* different variable, eagerly rename and continue *)
                        let val (cx, c) = alpha_vary (cx, c) replace_con_con
                            (* now, cx \not\in fv(confor) *)
                            val c = subst_con_con confor convar c
                        in
                            (cx, c)
                        end)
                incon

    fun subst_con_exp con convar exp =
        map_exp (fn x => ExVar x)
                (fn fbnds =>
                    map (fn (f, x, c1, c2, e) =>
                            (f, x, subst_con_con con convar c1,
                                   subst_con_con con convar c2,
                                   subst_con_exp con convar e))
                        fbnds)
                (fn (ex, e) => (ex, subst_con_exp con convar e))
                (fn (cx, e) =>
                    if Variable.eq (cx, convar) then
                        (cx, e)
                    else
                        let val (cx, e) = alpha_vary (cx, e) replace_con_exp
                            val e = subst_con_exp con convar e
                        in
                            (cx, e)
                        end)
                (subst_con_con con convar)
                exp

    fun subst_con_cons confor convar consin =
        map (subst_con_con confor convar) consin
  end
