structure IL3Print =
  struct
    val $ = Layout.str
    val % = Layout.mayAlign
    val & = Layout.seq


    open IL3

    fun var v = $(Variable.toString v)

    (*** kinds ***)

    fun atomic_kind_to_layout knd =
        case knd of
            KndType => $"Type"
          | KndUnit => $"1"
          | _ => Layout.paren (kind_to_layout knd)

    and prod_kind_to_layout knd =
        case knd of
            KndPair (k1, k2) =>
                %[atomic_kind_to_layout k1, $"*", atomic_kind_to_layout k2]
          | _ => atomic_kind_to_layout knd

    and kind_to_layout knd =
        case knd of
            KndArrow (k, k') =>
                %[prod_kind_to_layout k, $"->", kind_to_layout k']
          | _ => prod_kind_to_layout knd


    (*** constructors ***)

    fun bracket_con c = &[$"[", con_to_layout c, $"]"]

    and atomic_con_to_layout con =
        case con of
            ConVar cx => var cx
          | TyLit lt =>
                (case lt of
                    Literal.LitTyInt => $"int"
                  | Literal.LitTyChar => $"char"
                  | Literal.LitTyString => $"string")
          | TyTagged => $"tagged"
          | TyProd cons =>
                &[$"*",
                  Layout.sequence ("[", "]", ",") (map con_to_layout cons)]
          | TySum cons =>
                &[$"+",
                  Layout.sequence ("[", "]", ",") (map con_to_layout cons)]
          | ConUnit => $"<>"
          | ConPair (c1, c2) =>
                Layout.sequence ("<", ">", ",")
                    (map con_to_layout [c1, c2])
          | _ => Layout.paren (con_to_layout con)

    and app_con_to_layout con =
        case con of
            TyRef conarg =>
                &[$"ref ", atomic_con_to_layout conarg]
          | TyTag conarg =>
                &[$"tag ", atomic_con_to_layout conarg]
          | ConApp (c1, c2) =>
                %[app_con_to_layout c1,
                  atomic_con_to_layout c2]
          | ConProj1 c => %[ $"#1", atomic_con_to_layout c]
          | ConProj2 c => %[ $"#2", atomic_con_to_layout c]
          | _ => atomic_con_to_layout con

    and arrow_con_to_layout con =
        case con of
            TyArrow (con1, con2) =>
                %[app_con_to_layout con1,
                  &[$"-> ", arrow_con_to_layout con2]]
          | _ => app_con_to_layout con

    and con_to_layout con =
        (* first, eliminate bound evars *)
        case con of
            TyForall (cx, k, c) =>
                %[ &[$"All ", var cx, $":", atomic_kind_to_layout k, $"."],
                   con_to_layout c ]
          | ConAbs (cx, k, c) =>
                %[ &[$"\\ ", var cx, $":", atomic_kind_to_layout k, $"."],
                   con_to_layout c ]
          | ConMu (cx, k, c) =>
                %[ &[$"mu ", var cx, $":", atomic_kind_to_layout k, $"."],
                   con_to_layout c ]
          | _ => arrow_con_to_layout con


    (*** expressions ***)

    and atomic_exp_to_layout exp =
        case exp of
            ExVar x => var x
          | ExLit lit =>
                (case lit of
                    Literal.LitInt i => $(Int.toString i)
                  | Literal.LitChar c => $("'" ^ Char.toString c ^ "'")
                  | Literal.LitString s => $("\"" ^ s ^ "\""))
          | ExProd exps =>
                Layout.tuple (map exp_to_layout exps)
          | ExFix (i, fbnds) =>
                let fun fbnds_to_layout fbnds =
                    map (fn (f, x, c1, c2, e) =>
                        %[ &[var f, $" (", var x, $" : ", con_to_layout c1,
                                   $") : ", con_to_layout c2, $" |->" ],
                           Layout.indent 4 (exp_to_layout e) ])
                        fbnds
                in
                    %[&[$"fix[", $(Int.toString i), $"]"],
                      Layout.indent 4 (%(fbnds_to_layout fbnds)),
                      $"end"]
                end
          | _ => Layout.paren (exp_to_layout exp)

    and app_exp_to_layout exp =
        case exp of
            ExApp (e1, e2) => %[app_exp_to_layout e1,
                                atomic_exp_to_layout e2]
          | ExConApp (e, c) => %[app_exp_to_layout e,
                                 bracket_con c]
          | ExProj (i, e) => %[ &[$"#", $(Int.toString i)],
                                atomic_exp_to_layout e]
          | ExRaise (c, e) =>
                %[ $"raise", Layout.tuple [con_to_layout c,
                                            exp_to_layout e] ]
          | ExRef e =>
                %[ $"ref", atomic_exp_to_layout e ]
          | ExGet e =>
                %[ $"get", atomic_exp_to_layout e ]
          | ExSet (e1, e2) =>
                %[ $"set", Layout.tuple [exp_to_layout e1,
                                          exp_to_layout e2] ]
          | ExRoll (c, e) =>
                %[ $"roll", bracket_con c, atomic_exp_to_layout e ]
          | ExUnroll e =>
                %[ $"unroll", atomic_exp_to_layout e ]
          | ExSumInj (c, i, e) =>
                %[ $"inj", bracket_con c, $(Int.toString i),
                   atomic_exp_to_layout e ]
          | ExTagInj (e1, e2) =>
                %[ $"tag", Layout.tuple [exp_to_layout e1,
                                         exp_to_layout e2] ]
          | ExNewTag c =>
                %[ $"newtag", bracket_con c ]
          | ExUnOp (oper, e) =>
                %[ $(Literal.unop_to_string oper), atomic_exp_to_layout e]
          | _ => atomic_exp_to_layout exp

    and times_exp_to_layout exp =
        case exp of
            ExBinOp (e1, Literal.OpTimes, e2) =>
                %[times_exp_to_layout e1,
                  &[$"* ", app_exp_to_layout e2]]
          | _ => app_exp_to_layout exp
    
    and plus_minus_exp_to_layout exp =
        case exp of
            ExBinOp (e1, Literal.OpPlus, e2) =>
                %[plus_minus_exp_to_layout e1,
                  &[$"+ ", times_exp_to_layout e2]]
          | ExBinOp (e1, Literal.OpMinus, e2) =>
                %[plus_minus_exp_to_layout e1,
                  &[$"- ", times_exp_to_layout e2]]
          | ExBinOp (e1, Literal.OpConcat, e2) =>
                %[plus_minus_exp_to_layout e1,
                  &[$"^ ", times_exp_to_layout e2]]
          | _ => times_exp_to_layout exp

    and equals_exp_to_layout exp =
        case exp of
            ExBinOp (e1, Literal.OpEquals, e2) =>
                %[plus_minus_exp_to_layout e1,
                  &[$"= ", plus_minus_exp_to_layout e2]]
          | _ => plus_minus_exp_to_layout exp

    and exp_to_layout exp =
        case exp of
            ExIfTagOf (e1, e2, e3, e4) =>
                Layout.align
                [ &[$"iftagof ", exp_to_layout e1, $" is ",
                                 exp_to_layout e2, $" then"],
                  Layout.indent 4 (exp_to_layout e3),
                  $"else",
                  Layout.indent 4 (exp_to_layout e4) ]
          | ExCase (c, e, exps) =>
                Layout.align
                [ &[$"case ", bracket_con c, $" ", exp_to_layout e, $" of"],
                  Layout.indent 4 (Layout.sequence ("{", "}", ",")
                                    (map exp_to_layout exps)) ]
          | ExConAbs (cx, k, e) =>
                %[ &[$"/\\ ", var cx, $":", atomic_kind_to_layout k, $"."],
                   exp_to_layout e]
          | ExHandle (e1, (x, e2)) =>
                %[ $"handle",
                   Layout.indent 4 (exp_to_layout e1),
                   $"with",
                   Layout.indent 4 (&[var x, $" |-> ", exp_to_layout e2]) ]
          | ExLet (x, e1, e2) =>
                Layout.align
                [ %[ &[$"let ", var x, $" ="],
                     Layout.indent 4 (exp_to_layout e1) ],
                  $"in",
                  (exp_to_layout e2) ]
          | _ => equals_exp_to_layout exp

  end
