structure IL4Print =
struct
  open IL4

  val $ = Layout.str
  val % = Layout.mayAlign
  val & = Layout.seq

  fun var v = $(Variable.toString v)

  (*** kinds ***)

  fun atomic_kind_to_layout knd =
      case knd of
          KndType => $"Type"
        | KndUnit => $"1"
        | _ => Layout.paren (kind_to_layout knd)

  and prod_kind_to_layout knd =
      case knd of
          KndPair (k1, k2) =>
              %[atomic_kind_to_layout k1, $"*", atomic_kind_to_layout k2]
        | _ => atomic_kind_to_layout knd

  and kind_to_layout knd =
      case knd of
          KndArrow (k, k') =>
              %[prod_kind_to_layout k, $"->", kind_to_layout k']
        | _ => prod_kind_to_layout knd


  (*** constructors ***)

  fun bracket_con c = &[$"[", con_to_layout c, $"]"]

  and atomic_con_to_layout con =
      case con of
          ConVar cx => var cx
        | TyLit lt =>
              (case lt of
                  Literal.LitTyInt => $"int"
                | Literal.LitTyChar => $"char"
                | Literal.LitTyString => $"string")
        | TyTagged => $"tagged"
        | TyProd cons =>
              &[$"*",
                Layout.sequence ("[", "]", ",") (map con_to_layout cons)]
        | TySum cons =>
              &[$"+",
                Layout.sequence ("[", "]", ",") (map con_to_layout cons)]
        | ConUnit => $"<>"
        | ConPair (c1, c2) =>
              Layout.sequence ("<", ">", ",")
                  (map con_to_layout [c1, c2])
        | _ => Layout.paren (con_to_layout con)

  and cont_con_to_layout con =
      case con of
          TyCont c => &[$"~", cont_con_to_layout c]
        | _ => atomic_con_to_layout con

  and app_con_to_layout con =
      case con of
          TyRef conarg =>
              &[$"ref ", atomic_con_to_layout conarg]
        | TyTag conarg =>
              &[$"tag ", atomic_con_to_layout conarg]
        | ConApp (c1, c2) =>
              %[app_con_to_layout c1,
                atomic_con_to_layout c2]
        | ConProj1 c => %[ $"#1", atomic_con_to_layout c]
        | ConProj2 c => %[ $"#2", atomic_con_to_layout c]
        | _ => cont_con_to_layout con

  and con_to_layout con =
      (* first, eliminate bound evars *)
      case con of
          TyForall (cx, k, c) =>
              %[ &[$"All ", var cx, $":", atomic_kind_to_layout k, $"."],
                 con_to_layout c ]
        | ConAbs (cx, k, c) =>
              %[ &[$"\\ ", var cx, $":", atomic_kind_to_layout k, $"."],
                 con_to_layout c ]
        | ConMu (cx, k, c) =>
              %[ &[$"mu ", var cx, $":", atomic_kind_to_layout k, $"."],
                 con_to_layout c ]
        | _ => app_con_to_layout con


  (*** values ***)
  fun atomic_value_to_layout value =
      case value of
          ValVar x => var x
        | ValLit lit =>
              (case lit of
                  Literal.LitInt i => $(Int.toString i)
                | Literal.LitChar c => $("'" ^ Char.toString c ^ "'")
                | Literal.LitString s => $("\"" ^ s ^ "\""))
        | ValProd exps =>
              Layout.tuple (map value_to_layout exps)
        | ValFix (i, fbnds) =>
              let fun fbnds_to_layout fbnds =
                  map (fn (f, x, c, e) =>
                      %[ &[var f, $" (", var x, $" : ", con_to_layout c,
                                 $") |->" ],
                         Layout.indent 4 (exp_to_layout e) ])
                      fbnds
              in
                  %[&[$"fix[", $(Int.toString i), $"]"],
                    Layout.indent 4 (%(fbnds_to_layout fbnds)),
                    $"end"]
              end
        | _ => Layout.paren (value_to_layout value)

  and app_value_to_layout value =
      case value of
          ValRoll (c, v) =>
              %[ $"roll", bracket_con c, atomic_value_to_layout v ]
        | ValSumInj (c, i, v) =>
              %[ $"inj", bracket_con c, $(Int.toString i),
                 atomic_value_to_layout v ]
        | ValTagInj (v1, v2) =>
              %[ $"tag", Layout.tuple [value_to_layout v1,
                                       value_to_layout v2] ]
        | ValConApp (v, c) => %[app_value_to_layout v,
                                bracket_con c]
        | _ => atomic_value_to_layout value

  and value_to_layout value =
      case value of
          ValConAbs (cx, k, v) =>
              %[ &[$"/\\ ", var cx, $":", atomic_kind_to_layout k, $"."],
                 value_to_layout v]
        | _ => app_value_to_layout value

  (*** expressions ***)

  and atomic_exp_to_layout exp =
      case exp of
          _ => Layout.paren (exp_to_layout exp)

  and app_exp_to_layout exp =
      case exp of
          ExHalt v => %[$"halt", atomic_value_to_layout v]
        | ExApp (v1, v2) => %[app_value_to_layout v1,
                              atomic_value_to_layout v2]
        | _ => atomic_exp_to_layout exp

  and exp_to_layout exp =
      let fun arm_to_layout (x, e) = &[var x, $" |-> ", exp_to_layout e]
      in
          case exp of
              ExIfTagOf (v1, v2, (x, e3), e4) =>
                  Layout.align
                  [ &[$"iftagof ", value_to_layout v1, $" is ",
                                   value_to_layout v2, $" then"],
                    Layout.indent 4 (arm_to_layout (x, e3)),
                    $"else",
                    Layout.indent 4 (exp_to_layout e4) ]
            | ExCase (v, xexps) =>
                  Layout.align
                  [ &[$"case ", value_to_layout v, $" of"],
                    Layout.indent 4 (Layout.sequence ("{", "}", ",")
                                      (map arm_to_layout xexps)) ]
            | ExLet (b, e) =>
                  %[ &[$"let ", bnd_to_layout b],
                     $"in",
                     (exp_to_layout e) ]
            | _ => app_exp_to_layout exp
      end

  and bnd_to_layout bnd =
      let fun layb (x, l) = %[ &[var x, $" ="], Layout.indent 4 l ]
      in
          case bnd of
              BndVal (x, v) => layb (x, value_to_layout v)
            | BndProj (x, i, v) =>
                  layb (x, %[ &[$"#", $(Int.toString i)],
                              atomic_value_to_layout v])
            | BndUnOp (x, oper, v) =>
                  layb (x, %[ $(Literal.unop_to_string oper),
                               value_to_layout v])
            | BndBinOp (x, v1, oper, v2) =>
                  layb (x, %[value_to_layout v1,
                             &[$(Literal.binop_to_string oper), $" ",
                               value_to_layout v2]])
            | BndRef (x, v) =>
                  layb (x, %[ $"ref", atomic_value_to_layout v ])
            | BndGet (x, v) =>
                  layb (x, %[ $"get", atomic_value_to_layout v ])
            | BndSet (x, v1, v2) =>
                  layb (x, %[ $"set", Layout.tuple [value_to_layout v1,
                                                    value_to_layout v2] ])
            | BndUnroll (x, v) =>
                  layb (x, %[ $"unroll", atomic_value_to_layout v ])
            | BndNewTag (x, c) =>
                  layb (x, %[ $"newtag", bracket_con c ])
      end
end
