structure IL6 :> IL6 =
  struct
    type expvar = Variables.exp Variable.var
    type convar = Variables.con Variable.var
    type modvar = Variables.mod Variable.var
    
    datatype kind = datatype IL5.kind
    
    datatype con =
        ConVar of convar
      | ConAbs of convar * kind * con
      | ConApp of con * con
      | ConUnit
      | ConPair of con * con
      | ConProj1 of con
      | ConProj2 of con
      | ConMu of convar * kind * con (* convar bound in con, kind = KndType^n *)

      | TyLit of Literal.lit_ty
      | TyCont of (convar * kind) list * con (* instead of TyCont *)
      | TyRef of con
      | TyTagged
      | TyTag of con
      | TySum of con list
      | TyProd of con list
      | TyForall of convar * kind * con (* convar bound in con *)
      | TyExists of convar * kind * con
    
    datatype word = 
        WProd of value list
      | WRoll of con * value
      | WSumInj of con * int * value
      | WPack of ((con * value) * con)
      | WTagInj of value * value
      | WRef of value
      | WNewTag of con

    and value =
        ValVar of expvar
      | ValLit of Literal.lit
      | ValConAbs of convar * kind * value
      | ValConApp of value * con
      | ValLet of bnd * value
      
    and exp =
        ExHalt of value
      | ExApp of value * value
      | ExCase of value * (expvar * exp) list
      | ExIfTagOf of value * value * (expvar * exp) * exp
                                   (* expvar bound in exp1 *)
      | ExLet of bnd * exp (* bnd's expvar bound in exp2 *)
      | ExUnpack of convar * expvar * value * exp (* vars bound in exp *)
    
    and bnd =
        BndVal of expvar * value
      | BndAlloc of expvar * word
      | BndUnOp of expvar * Literal.unop * value
      | BndBinOp of expvar * value * Literal.binop * value
      | BndGet of expvar * value
      | BndSet of expvar * value * value
      | BndProj of expvar * int * value
      | BndUnroll of expvar * value
    type fbnd = ((convar * kind) list * expvar * con * exp)
                    (* Lam [a : k].lam x : c. e *)

    type program = (expvar * fbnd) list * exp

    (*** derived forms ***)

    fun kndProd ks = foldr KndPair KndUnit ks

    fun conProd cs = foldr ConPair ConUnit cs

    fun conProj (n, c) =
        if n = 1
        then ConProj1 c
        else conProj (n-1, ConProj2 c)

    val tyUnit = TyProd []
    val tyBool = TySum [tyUnit, tyUnit]

    val tyInt = TyLit Literal.LitTyInt
    val tyChar = TyLit Literal.LitTyChar
    val tyString = TyLit Literal.LitTyString

    local
      val u = Variable.newNamed "u"
    in
      val valUnit = ValLet(BndAlloc(u, WProd []), ValVar u)
    end

    fun valInt n = ValLit (Literal.LitInt n)
    fun valChar c = ValLit (Literal.LitChar c)
    fun valString s = ValLit (Literal.LitString s)

    (* mapping *)
    fun map_con fconvar fconvarcon con =
        let fun fcon con = map_con fconvar fconvarcon con
            fun fconvarkindcon (x, k, c) =
                    let val (x, c) = fconvarcon (x, c)
                    in
                        (x, k, c)
                    end
        in
            case con of
                ConVar x => fconvar x
              | ConAbs xkc => ConAbs (fconvarkindcon xkc)
              | ConApp (c1, c2) => ConApp (fcon c1, fcon c2)
              | ConUnit => ConUnit
              | ConPair (c1, c2) => ConPair (fcon c1, fcon c2)
              | ConProj1 c => ConProj1 (fcon c)
              | ConProj2 c => ConProj2 (fcon c)
              | ConMu xkc => ConMu (fconvarkindcon xkc)
              | TyLit lt => con
              | TyCont(vars, c) =>
                let
                  fun elemmapper ((a, k),(vars, c)) =
                    let val (a', k', c') = fconvarkindcon (a, k, c)
                    in ((a', k')::vars, c')
                    end
                in
                  TyCont(foldr elemmapper ([], fcon c) vars)
                end
              | TyRef c => TyRef (fcon c)
              | TyTagged => con
              | TyTag c => TyTag (fcon c)
              | TySum cons => TySum (map fcon cons)
              | TyProd cons => TyProd (map fcon cons)
              | TyForall xkc => TyForall (fconvarkindcon xkc)
              | TyExists xkc => TyExists (fconvarkindcon xkc)
        end

    (* map_word : (value -> value)
               -> (con -> con)
               -> word -> word *)
     fun map_word fvalue fcon word =
        case word of
          WProd vs => WProd(map fvalue vs)
        | WRoll(c, v) => WRoll(fcon c, fvalue v)
        | WSumInj(c, i, v) => WSumInj(fcon c, i, fvalue v)
        | WPack((c1, v), c2) => WPack((fcon c1, fvalue v), fcon c2)
        | WTagInj(v1, v2) => WTagInj(fvalue v1, fvalue v2)
        | WRef v => WRef(fvalue v)
        | WNewTag c => WNewTag(fcon c)
    
    (* map_value : (expvar -> value)
                -> (convar * value -> convar * value)
                -> (bnd * value -> bnd * value)
                -> (con -> con)
                -> value -> value *)
    and map_value fexpvar
                  fconvarvalue
                  fbndvalue
                  fcon
                  value =
        let val fvalue = map_value fexpvar fconvarvalue fbndvalue fcon
            fun fconvarkindvalue (x, k, e) =
                let val (x, e) = fconvarvalue (x, e)
                in
                    (x, k, e)
                end
        in
            case value of
                ValLit _ => value
              | ValVar x => fexpvar x
              | ValConAbs xkv => ValConAbs (fconvarkindvalue xkv)
              | ValConApp (v, c) => ValConApp (fvalue v, fcon c)
              | ValLet(b, v) => ValLet(fbndvalue (b, v))
        end
    
    (* map_exp : (expvar * exp -> expvar * exp)
              -> (convar * expvar * exp -> convar * expvar * exp
              -> (bnd * exp -> bnd * exp)
              -> (value -> value)
              -> exp -> exp *)
    fun map_exp fexpvarexp
                fconexpvarexp
                fbndexp
                fvalue
                exp =
        let val fexp = map_exp fexpvarexp fconexpvarexp fbndexp fvalue
        in
            case exp of
                ExHalt v => ExHalt (fvalue v)
              | ExApp (v1, v2) => ExApp (fvalue v1, fvalue v2)
              | ExCase (v, xes) => ExCase (fvalue v, map fexpvarexp xes)
              | ExIfTagOf (v1, v2, xe1, e2) =>
                    ExIfTagOf (fvalue v1, fvalue v2, fexpvarexp xe1, fexp e2)
              | ExLet bndexp => ExLet (fbndexp bndexp)
              | ExUnpack (a, x, v, e) =>
                  let
                    val (a', x', e') = fconexpvarexp (a, x, e)
                    val v' = fvalue v
                  in
                    ExUnpack(a', x', v', e')
                  end
        end
    
    (* map_bnd : (value -> value)
              -> (word -> word)
              -> (con -> con)
              -> bnd -> bnd *)
    fun map_bnd fvalue fword fcon bnd =
        let val fbnd = map_bnd fvalue fword fcon
        in
            case bnd of
                BndVal (x, v) => BndVal (x, fvalue v)
              | BndAlloc (x, w) => BndAlloc(x, fword w)
              | BndUnOp (x, oper, v) => BndUnOp(x, oper, fvalue v)
              | BndBinOp (x, v1, oper, v2) =>
                  BndBinOp (x, fvalue v1, oper, fvalue v2)
              | BndGet (x, v) => BndGet (x, fvalue v)
              | BndSet (x, v1, v2) => BndSet (x, fvalue v1, fvalue v2)
              | BndUnroll (x, v) => BndUnroll (x, fvalue v)
              | BndProj (x, i, v) => BndProj (x, i, fvalue v)
        end
    
    fun bnd_var bnd =
        case bnd of
            BndVal (x, _) => x
          | BndUnOp (x, _, _) => x
          | BndBinOp (x, _, _, _) => x
          | BndAlloc (x, _) => x
          | BndGet (x, _) => x
          | BndSet (x, _, _) => x
          | BndUnroll (x, _) => x
          | BndProj (x, _, _) => x
    
    (*** replacement of convars ***)

    fun replace_con_con convar convar' con =
        map_con (fn cx =>
                    if Variable.eq (cx, convar) then
                        ConVar convar'
                    else
                        ConVar cx)
                (fn (cx, c) =>
                    if Variable.eq (cx, convar) then
                        (cx, c)
                    else
                        (cx, replace_con_con convar convar' c))
                con
    
    fun replace_con_value convar convar' value =
        map_value (fn x => ValVar x)
                  (fn (cx, v) =>
                    if Variable.eq (cx, convar) then
                        (cx, v)
                    else
                        (cx, replace_con_value convar convar' v))
                  (fn (b, v) => (replace_con_bnd convar convar' b,
                                 replace_con_value convar convar' v))
                  (replace_con_con convar convar')
                  value

    and replace_con_exp convar convar' exp =
        map_exp (fn (ex, e) => (ex, replace_con_exp convar convar' e))
                (fn (cx, ex, e) =>
                  if Variable.eq (cx, convar) then
                    (cx, ex, e)
                  else
                    (cx, ex, replace_con_exp convar convar' e))
                (fn (b, e) => (replace_con_bnd convar convar' b,
                               replace_con_exp convar convar' e))
                (replace_con_value convar convar')
                exp
    
    and replace_con_word convar convar' word =
        map_word (replace_con_value convar convar')
                 (replace_con_con convar convar')
                 word

    and replace_con_bnd convar convar' bnd =
        map_bnd (replace_con_value convar convar')
                (replace_con_word convar convar')
                (replace_con_con convar convar')
                bnd
    
    (*** replacement of expvars ***)

    fun replace_exp_value expvar expvar' value =
        map_value (fn x =>
                    if Variable.eq (x, expvar) then
                        ValVar expvar'
                    else
                        ValVar x)
                  (fn (cx, v) => (cx, replace_exp_value expvar expvar' v))
                  (fn (b, v) =>
                      let val b = replace_exp_bnd expvar expvar' b
                          val x = bnd_var b
                      in
                          if Variable.eq (x, expvar) then
                              (b, v)
                          else
                              (b, replace_exp_value expvar expvar' v)
                      end)
                  (fn c => c) (* expvars do not appear in cons *)
                  value

    and replace_exp_bnd expvar expvar' bnd =
        map_bnd (replace_exp_value expvar expvar')
                (replace_exp_word expvar expvar')
                (fn c => c) (* expvars do not appear in cons *)
                bnd
    
    and replace_exp_word expvar expvar' word =
        map_word (replace_exp_value expvar expvar')
                 (fn c => c) (* expvars do not appear in cons *)
                 word

    and replace_exp_exp expvar expvar' exp =
        map_exp (fn (ex, e) =>
                    if Variable.eq (ex, expvar) then
                        (ex, e)
                    else
                        (ex, replace_exp_exp expvar expvar' e))
                (fn (cx, ex, e) =>
                    if Variable.eq (ex, expvar) then
                        (cx, ex, e)
                    else
                        (cx, ex, replace_exp_exp expvar expvar' e))
                (fn (b, e) =>
                    let val b = replace_exp_bnd expvar expvar' b
                        val x = bnd_var b
                    in
                        if Variable.eq (x, expvar) then
                            (b, e)
                        else
                            (b, replace_exp_exp expvar expvar' e)
                    end)
                (replace_exp_value expvar expvar')
                exp
    
    fun alpha_vary (c1var, c2obj) replace =
        let val c1var' = Variable.freshen c1var
            val c2obj' = replace c1var c1var' c2obj
        in
            (c1var', c2obj')
        end

    fun alpha_vary2 (varA, objA) (varB, objB) replace =
        let val var' = Variable.freshen varA
            val objA' = replace varA var' objA
            val objB' = replace varB var' objB
        in
            (var', (objA', objB'))
        end
    
    fun alpha_vary_bnd_exp (bnd, exp) =
        case bnd of
            BndVal (x, v) =>
                let val (x, exp) = alpha_vary (x, exp) replace_exp_exp
                in
                    (BndVal (x, v), exp)
                end
          | BndAlloc(x, w) =>
                let val (x, exp) = alpha_vary (x, exp) replace_exp_exp
                in
                    (BndAlloc (x, w), exp)
                end
          | BndUnOp (x, oper, v) =>
                let val (x, exp) = alpha_vary (x, exp) replace_exp_exp
                in
                    (BndUnOp (x, oper, v), exp)
                end
          | BndBinOp (x, v1, oper, v2) =>
                let val (x, exp) = alpha_vary (x, exp) replace_exp_exp
                in
                    (BndBinOp (x, v1, oper, v2), exp)
                end
          | BndGet (x, v) =>
                let val (x, exp) = alpha_vary (x, exp) replace_exp_exp
                in
                    (BndGet (x, v), exp)
                end
          | BndSet (x, v1, v2) =>
                let val (x, exp) = alpha_vary (x, exp) replace_exp_exp
                in
                    (BndSet (x, v1, v2), exp)
                end
          | BndUnroll (x, v) =>
                let val (x, exp) = alpha_vary (x, exp) replace_exp_exp
                in
                    (BndUnroll (x, v), exp)
                end
          | BndProj (x, i, v) =>
                let val (x, exp) = alpha_vary (x, exp) replace_exp_exp
                in
                    (BndProj (x, i, v), exp)
                end
    
    (*** substitution for convars ***)

    fun subst_con_con confor convar incon =
        map_con (fn cx =>
                    if Variable.eq (cx, convar) then
                        (* cx is convar free, substitute for it *)
                        confor
                    else
                        (* not convar -- put it back *)
                        ConVar cx)
                (fn (cx, c) =>
                    if Variable.eq (cx, convar) then
                        (* cx is convar being rebound, leave it alone *)
                        (cx, c)
                    else
                        (* different variable, eagerly rename and continue *)
                        let val (cx, c) = alpha_vary (cx, c) replace_con_con
                            (* now, cx \not\in fv(confor) *)
                            val c = subst_con_con confor convar c
                        in
                            (cx, c)
                        end)
                incon
  end
