
(* Combinator-style tokenizer. Takes characters as inputs, outputs
   items of type Tokens.token (see tokens.sml). This was adapted
   from a very old tokenizer for something else, so there are some
   aspects of it that are a little strange. *)

structure Tokenize :> TOKENIZE =
struct

  open Tokens

  open Parsing

  infixr 4 << >>
  infixr 3 &&
  infix  2 -- ##
  infix  2 wth suchthat return guard
  infixr 1 ||

  exception Lex of Pos.pos * string

  infix 2 check
  fun p check (pred, err) =
      !!p wth (fn (x, pos) => if pred x then x else raise Lex (pos, err x))

  val quotc = #"\"" (* " *)

  (* Report an error with its position. *)
  (* XXX should raise exception? *)
  (* fun error s pos = print (Pos.toString pos ^ ": " ^ s ^ "\n") *)
  fun error s pos = raise Lex (pos, s)

  (* Succeeds if p succeeds at the current point, but doesn't
     consume anything. *)

  fun ahead p = lookahead p (fn _ => succeed ())

  (* Ignores the result of p *)

  fun ignore p = p return ()

  fun midchar #"_" = true
    | midchar #"'" = true
    | midchar _ = false

  val isSymbolic = Char.contains "!%&$#+-/:<=>@\\~`^|*"

  (* these never appear as part of a longer identifier *)
  val isSep = Char.contains "()[],{};. "

  fun identchar c = Char.isAlpha c orelse Char.isDigit c orelse midchar c

  (* probably should have \r, \t, \x05, etc. *)
  val escapechar = 
    ((literal #"\\" >> literal quotc) ||
     (literal #"\\" >> literal #"\\") ||
     (literal #"\\" && literal #"n" return #"\n"))

  val getchar = (satisfy (fn x => x <> quotc andalso x <> #"\\") ||
                 escapechar)

  val integer = 
      (repeat1 (satisfy Char.isDigit))
      wth (Option.valOf o Int.fromString o implode)

  val insidechars = (repeat getchar) wth implode

  val stringlit = 
    (literal quotc) >>
    ((insidechars << (literal quotc)) 
      guard error "Unclosed string or bad escape.")

  (*
  val char = ((literal #"?" >> (satisfy (fn x => x <> #"\\"))) ||
              (literal #"?" >> escapechar))
  *)

  val char = string [#"#", quotc] >> getchar << literal quotc
  (*
  val char = literal #"#" >> stringlit
                check (fn s => size s = 1,
                       fn _ => "character constant not length 1")
                wth (fn s => String.sub (s, 0))
  *)

  val float = ((repeat  (satisfy Char.isDigit)) <<
               (literal #".")) &&
                  (repeat1 (satisfy Char.isDigit))
                  wth (fn (a, b) =>
                       (Option.valOf (Real.fromString 
                                      ("0." ^ (implode b)))) +
                       (case a of 
                            nil => 0.0
                          | _ => Real.fromInt 
                                (Option.valOf (Int.fromString 
                                               (implode a)))))

  val number = integer || (literal #"~" >> integer) wth op~

  val tyvar = literal #"'" && repeat (satisfy identchar) wth (implode o op::)

  (*
  fun comment () = string [#"(",#"*"]
      && (repeat ($insideComment) && string [#"*",#")"]
          guard error "Unterminated comment.")

  (* Either a nested comment or a single character (which is not
     start of a nested comment or the comment terminator). *)

  and insideComment () =
      ignore ($comment)
         || any -- (fn #"*" => ahead (satisfy (fn x => x <> #")"))
                     | #"(" => ahead (satisfy (fn x => x <> #"*"))
                     | _    => succeed ())
  *)

  exception Internal of string

  fun comment () = !!(string [#"(", #"*"]) -- (fn (_, p) => commentRest [p])

  and commentRest ps =
    alt [!!(string [#"(", #"*"]) -- (fn (_, p) => commentRest (p::ps)),
         string [#"*", #")"] -- (fn _ => case ps of
                                            [] => raise Internal "commentRest"
                                          | _::[] => succeed ()
                                          | _::ps => commentRest ps),
         any -- (fn _ => commentRest ps),
         done () -- (fn () => error "unterminated comment" (hd ps))]

  (* White space. *)

  val space = repeat (ignore ($comment) || ignore (satisfy Char.isSpace))

  val keywords =
      [("and", AND),
       ("andalso", ANDALSO),
       ("as", AS),
       ("case", CASE),
       ("datatype", DATATYPE),
       ("else", ELSE),
       ("end", END),
       ("exception", EXCEPTION),
       ("fn", FN),
       ("fun", FUN),
       ("functor", FUNCTOR),
       ("handle", HANDLE),
       ("if", IF),
       ("in", IN),
       ("include", INCLUDE),
       ("infix", INFIX),
       ("infixr", INFIXR),
       ("let", LET),
       ("local", LOCAL),
       ("nonfix", NONFIX),
       ("of", OF),
       ("op", OP),
       ("open", OPEN),
       ("orelse", ORELSE),
       ("sharing", SHARING),
       ("sig", SIG),
       ("signature", SIGNATURE),
       ("struct", STRUCT),
       ("structure", STRUCTURE),
       ("then", THEN),
       ("type", TYPE),
       ("val", VAL),
       ("where", WHERE),
       ("with", WITH),
       ("withtype", WITHTYPE),
       ("raise", RAISE),
       ("(", LPAREN),
       (")", RPAREN),
       ("{", LBRACE),
       ("}", RBRACE),
       ("[", LBRACKET),
       ("]", RBRACKET),
       (",", COMMA),
       (":", COLON),
       (":>", COLONGREATER),
       (";", SEMICOLON),
       ("_", UNDERSCORE),
       ("|", BAR),
       ("=", EQUALS),
       ("=>", DARROW),
       ("->", ARROW),
       ("#", HASH),
       (".", DOT) ]

  fun ident s =
      let
          fun id nil = IDENT s
            | id ((a,b)::t) = if a = s then b else id t
      in
          id keywords
      end

  val letters = satisfy Char.isAlpha && 
                repeat (satisfy identchar) wth op::

  val word = 
      alt [letters wth implode,
           (satisfy isSep) wth Char.toString,
           repeat1 (satisfy isSymbolic) wth implode]


  val token = space >> !! (alt [char wth CHAR,
                                float wth FLOAT,
                                number wth INT,
                                stringlit wth STRING,
                                tyvar wth TYVAR,
                                word wth ident,
                                literal #"_" return UNDERSCORE 
                                      (* XXX ok, but nasty... *)
                                ])
end
